# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" IPv6 Unicast """

import struct
import binascii
import netaddr

from yabgp.message.attribute.nlri import NLRI


class IPv6Unicast(NLRI):

    @classmethod
    def parse(cls, nlri_data):
        """
        decode IPv6 NLRI data
        :param nlri_data: NLRI raw hex data
        :return: return the results after decoding.
        """
        nlri_list = []
        while nlri_data:
            if isinstance(nlri_data[0], int):
                prefix_bit_len = int(nlri_data[0])
            else:
                prefix_bit_len = ord(nlri_data[0:1])
            if prefix_bit_len % 8 == 0:
                prefix_byte_len = prefix_bit_len / 8
            else:
                prefix_byte_len = prefix_bit_len / 8 + 1
            offset = int(prefix_byte_len + 1)
            prefix_addr = str(netaddr.IPAddress(int(binascii.b2a_hex(nlri_data[1:offset]), 16))) \
                + '/%s' % prefix_bit_len
            nlri_list.append(prefix_addr)
            nlri_data = nlri_data[int(prefix_byte_len) + 1:]

        return nlri_list

    @classmethod
    def construct(cls, nlri_list):
        """
        Construct NLRI from list to hex data
        :param nlri_list:
        :return:
        """
        nlri_hex = b''
        for prefix in nlri_list:
            prefix = netaddr.IPNetwork(prefix)
            nlri_hex += struct.pack('!B', prefix.prefixlen)
            nlri_hex += binascii.unhexlify(hex(prefix.ip)[2:])
        return nlri_hex



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.message.attribute.tunnelencaps import TunnelEncaps


class TestTunnelEncaps(unittest.TestCase):

    def test_construct_optional_label_sid(self):
        option_sid = 16384255
        data_dict = {
            "label": 4000,
            "TC": 0,
            "S": 0,
            "TTL": 255
        }
        self.assertEqual(option_sid, TunnelEncaps.construct_optional_label_sid(data_dict))

    def test_construct_weight(self):
        data_dict = {"9": 10, "1": []}
        segment_list = dict([(int(l), r) for (l, r) in data_dict.items()])
        weight_hex = b'\x09\x06\x00\x00\x00\x00\x00\x0a'
        sid_hex = b''
        self.assertEqual((weight_hex, sid_hex), TunnelEncaps.construct_weight_and_seg(segment_list))

    def test_construct_seg(self):
        data_dict = {
            "1": [
                {
                    "1": {
                        "label": 2000
                    }
                },
                {
                    "3": {
                        "node": "10.1.1.1",
                        "SID": {
                            "label": 2000,
                            "TC": 0,
                            "S": 0,
                            "TTL": 255
                        }
                    }
                }
            ]
        }
        segment_list = dict([(int(l), r) for (l, r) in data_dict.items()])
        weight_hex = b''
        sid_hex = b'\x01\x06\x00\x00\x00\x7d\x00\xff\x03\x0a\x00\x00\x0a\x01\x01\x01\x00\x7d\x00\xff'
        self.assertEqual((weight_hex, sid_hex), TunnelEncaps.construct_weight_and_seg(segment_list))

    def test_construct_old_binding_sid(self):
        data_dict = {"0": "old", "7": 25102}
        data_hex = b'\x07\x06\x00\x00\x06\x20\xe0\x00'
        self.assertEqual(data_hex, TunnelEncaps.construct(data_dict)[8:])

    def test_construct_old_preference(self):
        data_dict = {"0": "old", "6": 100}
        data_hex = b'\x06\x06\x00\x00\x00\x00\x00\x64'
        self.assertEqual(data_hex, TunnelEncaps.construct(data_dict)[8:16])

    def test_construct_new_binding_sid(self):
        data_dict = {"0": "new", "13": 25102}
        data_hex = b'\x0d\x06\x00\x00\x06\x20\xe0\x00'
        self.assertEqual(data_hex, TunnelEncaps.construct(data_dict)[8:])

    def test_construct_new_preference(self):
        data_dict = {"0": "new", "12": 100}
        data_hex = b'\x0c\x06\x00\x00\x00\x00\x00\x64'
        self.assertEqual(data_hex, TunnelEncaps.construct(data_dict)[8:16])

    def test_construct_segement_lists(self):
        data_dict = {
            "0": "old",
            "128": [
                {
                    "9": 10,
                    "1": [
                        {
                            "1": {
                                "label": 2000
                            }
                        },
                        {
                            "3": {
                                "node": "10.1.1.1",
                                "SID": {
                                    "label": 2000,
                                    "TC": 0,
                                    "S": 0,
                                    "TTL": 255
                                }
                            }
                        }
                    ]
                }
            ]
        }
        data_hex_1 = b'\x80\x00\x1d\x00\x09\x06\x00\x00\x00\x00\x00\x0a\x01\x06\x00\x00\x00\x7d\x00' \
                     b'\xff\x03\x0a\x00\x00\x0a\x01\x01\x01\x00\x7d\x00\xff\x07\x02\x00\x00'
        data_hex_2 = b'\x07\x02\x00\x00\x80\x00\x1d\x00\x09\x06\x00\x00\x00\x00\x00\x0a\x01\x06\x00' \
                     b'\x00\x00\x7d\x00\xff\x03\x0a\x00\x00\x0a\x01\x01\x01\x00\x7d\x00\xff'
        self.assertTrue(TunnelEncaps.construct(data_dict)[8:] in [data_hex_1, data_hex_2])

if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Unittest for MPUnReach NLRI"""

import unittest
from yabgp.message.attribute.mpunreachnlri import MpUnReachNLRI


class TestMpUnReachNLRI(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_ipv4_mpls_vpn_parse(self):
        data_bin = b'\x80\x0f\x12\x00\x01\x80\x70\x80\x00\x00\x00\x00\x00\x02\x00\x00\x00\x02\xc0\xa8\xc9'
        data_hoped = {'afi_safi': (1, 128),
                      'withdraw': [{'label': [524288],
                                    'rd': '2:2',
                                    'prefix': '192.168.201.0/24'}]}
        self.assertEqual(data_hoped, MpUnReachNLRI.parse(data_bin[3:]))

    def test_ipv4_mpls_vpn_construct(self):
        data_hoped = {'afi_safi': (1, 128),
                      'withdraw': [
                          {
                              'rd': '2:2',
                              'label': [524288],
                              'prefix': '192.168.201.0/24'}]}
        self.assertEqual(data_hoped, MpUnReachNLRI.parse(MpUnReachNLRI.construct(data_hoped)[4:]))

    def test_ipv6_unicast_parse(self):
        data_bin = b"\x00\x02\x01\x80\x20\x01\x48\x37\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20"
        data_hoped = {'afi_safi': (2, 1), 'withdraw': ['2001:4837::20/128']}
        self.assertEqual(data_hoped, MpUnReachNLRI.parse(data_bin))

    def test_ipv6_unicast_construct(self):
        nlri_dict = {
            'afi_safi': (2, 1),
            'withdraw': ['2001:3232::1/128', '::2001:3232:1:0/64', '2001:4837:1632::2/127']}
        self.assertEqual(nlri_dict, MpUnReachNLRI.parse(MpUnReachNLRI.construct(nlri_dict)[4:]))

    def test_ipv6_mpls_vpn_parse_construct(self):
        data_dict = {
            'afi_safi': (2, 128),
            'withdraw': [
                {'label': [524288], 'rd': '100:10', 'prefix': '2001:3232::1/128'},
                {'label': [524288], 'rd': '100:11', 'prefix': '2001:4837:1632::2/127'}
            ]
        }
        self.assertEqual(
            data_dict,
            MpUnReachNLRI.parse(
                value=MpUnReachNLRI.construct(value=data_dict)[4:]
            ))

    def test_ipv4_flowspec_parse(self):
        data_bin = b'\x00\x01\x85\x0a\x01\x18\xc0\x55\x02\x02\x18\xc0\x55\x01'
        nlri_dict = {'afi_safi': (1, 133), 'withdraw': [{1: '192.85.2.0/24', 2: '192.85.1.0/24'}]}
        self.assertEqual(nlri_dict, MpUnReachNLRI.parse(data_bin))

    def test_ipv4_flowspec_construct(self):
        data_bin = b'\x00\x01\x85\x0a\x01\x18\xc0\x55\x02\x02\x18\xc0\x55\x01'
        nlri_dict = {'afi_safi': (1, 133), 'withdraw': [{1: '192.85.2.0/24', 2: '192.85.1.0/24'}]}
        self.assertEqual(data_bin, MpUnReachNLRI.construct(nlri_dict)[4:])

    def test_ipv4_srte_construct(self):
        data_bin = b'\x00\x01\x49\x60\x00\x00\x00\x00\x00\x00\x00\x0a\xc0\xa8\x05\x07'
        nlri_dict = {'afi_safi': (1, 73), 'withdraw': {'distinguisher': 0, 'color': 10, 'endpoint': '192.168.5.7'}}
        self.assertEqual(data_bin, MpUnReachNLRI.construct(nlri_dict)[4:])

    def test_l2vpn_evpn_route_type_1_parse_construct(self):
        data_dict = {
            "afi_safi": (25, 70),
            "withdraw": [{
                "type": 1,
                "value": {
                    "rd": "1.1.1.1:32867",
                    "esi": 0,
                    "eth_tag_id": 100,
                    "label": [10]
                }
            }]}
        self.assertEqual(data_dict, MpUnReachNLRI.parse(MpUnReachNLRI.construct(data_dict)[4:]))

    def test_l2vpn_evpn_route_type_2_parse_construct(self):
        data_dict = {
            'afi_safi': (25, 70),
            'withdraw': [
                {
                    'type': 2,
                    'value': {
                        'eth_tag_id': 108,
                        'ip': '11.11.11.1',
                        'label': [0],
                        'rd': '172.17.0.3:2',
                        'mac': '00-11-22-33-44-55',
                        'esi': 0}}]
        }
        self.assertEqual(data_dict, MpUnReachNLRI.parse(MpUnReachNLRI.construct(data_dict)[4:]))

    def test_l2vpn_evpn_route_type_3_parse_construct(self):
        data_dict = {
            "afi_safi": (25, 70),
            "withdraw": [
                {
                    "type": 3,
                    "value": {
                        "rd": "172.16.0.1:5904",
                        "eth_tag_id": 100,
                        "ip": "192.168.0.1"
                    }
                }
            ]
        }
        self.assertEqual(data_dict, MpUnReachNLRI.parse(MpUnReachNLRI.construct(data_dict)[4:]))

    def test_l2vpn_evpn_route_type_4_parse_construct(self):
        data_dict = {
            "afi_safi": (25, 70),
            "withdraw": [
                {
                    "type": 4,
                    "value": {
                        "rd": "172.16.0.1:8888",
                        "esi": 0,
                        "ip": "192.168.0.1"
                    }
                }
            ]
        }
        self.assertEqual(data_dict, MpUnReachNLRI.parse(MpUnReachNLRI.construct(data_dict)[4:]))


if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/draft-ietf-idr-bgp-ls-segment-routing-ext-03#section-2.1.4
#     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |               Type            |               Length          |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |      Flags    |   RESERVED    |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |                  Range Size                   |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    //                SID/Label sub-TLV (variable)                 //
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


@LinkState.register()
class SRLB(TLV):

    TYPE = 1036  # https://tools.ietf.org/html/draft-ietf-idr-bgp-ls-segment-routing-ext-03#section-2.1.4
    TYPE_STR = "srlb"

    @classmethod
    def unpack(cls, value):
        """
        """
        value = value[2:]
        results = []
        while True:
            if len(value) == 0:
                break
            else:
                tmp = dict()
                range_size = struct.unpack('!I', "\x00" + value[:3])[0]
                tmp['range_size'] = range_size
                length = struct.unpack('!H', value[5:7])[0]
                if length == 3:
                    data = (struct.unpack('!I', "\x00" + value[7:7 + length])[0] << 12) >> 12
                    value = value[7 + length:]
                    tmp['label'] = data
                elif length == 4:
                    data = struct.unpack('!I', value[7:7 + length])[0]
                    value = value[7 + length:]
                    tmp['sid'] = data
                results.append(tmp)
        return cls(value=results)



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii

import netaddr

from yabgp.common import afn
from yabgp.common import safn
from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri import NLRI


class MPLSVPN(NLRI):
    """
    IPv4/IPv6 MPLS VPN NLRI
    +---------------------------+
    | Length (1 octet)          |
    +---------------------------+
    | Label  (3 octet)          |
    +---------------------------+
    |...........................|
    +---------------------------+
    | Prefix (variable)         |
    +---------------------------+
    a) Length: The Length field indicates the length, in bits, of the address prefix.
    b) Label: (24 bits) Carries one or more labels in a stack, although a BGP update
    has only one label. This field carries the following parts of the MPLS shim header:
        Label Value-20 Bits
        Experimental bits-3 Bits
        Bottom of stack bit-1 bit
    c) Prefix: different coding way according to different SAFI
    Route Distinguisher (8 bytes) plus IPv4  prefix (32 bits) or IPv6 prefix(128 bits)
    rd (Route Distinguisher) structure (RFC 4364)
    """

    WITHDARW_LABEL_HEX = b'\x80\x00\x00'
    WITHDARW_LABEL = 524288
    AFI = None
    SAFI = None

    @classmethod
    def parse(cls, value, iswithdraw=False):
        """
        parse nlri
        :param value: the raw hex nlri value
        :param iswithdraw: if this is parsing withdraw
        :return: nlri list
        """
        nlri_list = []
        while value:
            nlri_dict = {}
            # for python2 and python3
            if isinstance(value[0], int):
                prefix_bit_len = value[0]
            else:
                prefix_bit_len = ord(value[0])
            if prefix_bit_len % 8 == 0:
                prefix_byte_len = int(prefix_bit_len / 8)
            else:
                prefix_byte_len = int(prefix_bit_len / 8) + 1

            if not iswithdraw:
                nlri_dict['label'] = cls.parse_mpls_label_stack(value[1:])
            else:
                nlri_dict['label'] = [MPLSVPN.WITHDARW_LABEL]

            nlri_dict['rd'] = MPLSVPN.parse_rd(value[4:12])
            prefix = value[12:prefix_byte_len + 1]
            if cls.AFI == afn.AFNUM_INET and cls.SAFI == safn.SAFNUM_LAB_VPNUNICAST:
                if len(prefix) < 4:
                    prefix += b'\x00' * (4 - len(prefix))
                nlri_dict['prefix'] = str(netaddr.IPAddress(struct.unpack('!I', prefix)[0])) +\
                    '/%s' % (prefix_bit_len - 88)
            elif cls.AFI == afn.AFNUM_INET6 and cls.SAFI == safn.SAFNUM_LAB_VPNUNICAST:
                if len(prefix) < 16:
                    prefix += b'\x00' * (16 - len(prefix))
                nlri_dict['prefix'] = str(netaddr.IPAddress(int(binascii.b2a_hex(prefix), 16))) +\
                    '/%s' % (prefix_bit_len - 88)
            value = value[prefix_byte_len + 1:]
            nlri_list.append(nlri_dict)
        return nlri_list

    @classmethod
    def construct(cls, value, iswithdraw=False):
        nlri_bin = b''
        for nlri in value:
            # construct label
            if iswithdraw:
                label_hex = MPLSVPN.WITHDARW_LABEL_HEX
            else:
                label_hex = MPLSVPN.construct_mpls_label_stack(nlri['label'])
            # construct rd
            rd_hex = MPLSVPN.construct_rd(nlri['rd'])
            # construct prefix
            prefix_str, prefix_len = nlri['prefix'].split('/')
            prefix_len = int(prefix_len)
            prefix_hex = b''
            if cls.AFI == afn.AFNUM_INET and cls.SAFI == safn.SAFNUM_LAB_VPNUNICAST:
                prefix_hex = cls.construct_prefix_v4(prefix_len, prefix_str)
            elif cls.AFI == afn.AFNUM_INET6 and cls.SAFI == safn.SAFNUM_LAB_VPNUNICAST:
                prefix_hex = cls.construct_prefix_v6(nlri['prefix'])
            prefix_hex = label_hex + rd_hex + prefix_hex
            prefix_len = struct.pack('!B', prefix_len + len(label_hex + rd_hex) * 8)
            nlri_bin += prefix_len + prefix_hex
        return nlri_bin

    @classmethod
    def parse_mpls_label_stack(cls, data):
        labels = []
        while len(data) >= 3:
            label = struct.unpack('!L', b'\00'+data[:3])[0]
            data = data[3:]
            labels.append(label >> 4)
            if label & 0x001:
                break
        return labels

    @classmethod
    def construct_mpls_label_stack(cls, labels):
        data = b''
        last_label = labels[-1]
        for label in labels[:-1]:
            data += struct.pack('!L', label << 4)[1:]
        if last_label != 0:
            data += struct.pack('!L', (last_label << 4 | 1))[1:]
        else:
            data += b'\x00\x00\x00'
        return data

    @classmethod
    def parse_rd(cls, data):
        """
        For Cisco: The BGP route distinguisher can be derived automatically
        from the VNI and BGP router ID of the VTEP switch
        :param data:
        :return:
        """
        rd_type = struct.unpack('!H', data[0:2])[0]
        rd_value = data[2:8]
        if rd_type == bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_0:
            asn, an = struct.unpack('!HI', rd_value)
            rd = '%s:%s' % (asn, an)
        elif rd_type == bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_1:
            ip = str(netaddr.IPAddress(struct.unpack('!I', rd_value[0:4])[0]))
            an = struct.unpack('!H', rd_value[4:6])[0]
            rd = '%s:%s' % (ip, an)
        elif rd_type == bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_2:
            asn, an = struct.unpack('!IH', rd_value)
            rd = '%s:%s' % (asn, an)
        else:
            # fixme(by xiaopeng163) for other rd type process
            rd = str(rd_value)
        return rd

    @classmethod
    def construct_rd(cls, data):
        # fixme(by xiaopeng163) for other rd type process
        data = data.split(':')
        if '.' in data[0]:
            return struct.pack('!H', bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_1) + netaddr.IPAddress(data[0]).packed + \
                struct.pack('!H', int(data[1]))
        else:
            data = [int(x) for x in data]
            if data[0] <= 0xffff:
                return struct.pack('!HHI', bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_0, data[0], data[1])
            else:
                return struct.pack('!HIH', bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_2, data[0], data[1])



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri.ipv4_flowspec import IPv4FlowSpec


class TestIPv4FlowSpec(unittest.TestCase):

    def test_parse(self):
        raw_hex = '\x01\x18\x02\x02\x02\x02\x10\x03\x03\x03\x01\x00\x01\x2f\x01\x58' \
                  '\x01\x01\x01\x02\x01\x59\x81\x67\x05\x11\x1f\x92\x91\x1f\x93\x06\x01' \
                  '\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93\x07\x01\x02\x01' \
                  '\x03\x01\x05\x81\x06\x08\x81\x02\x09\x81\x28\x0a\x01\xfe\x03\xfe\xd5' \
                  '\x01\x2c\x0b\x01\x28\x81\x30\x0c\x81\x01'
        self.assertEqual(
            {1: '2.2.2.0/24', 2: '3.3.0.0/16', 3: '=0|=47|=88|=1|=2|=89|=103', 5: '=8082|=8083',
             6: '=80|=8080|=8081|=8082|=8083', 7: '=2|=3|=5|=6', 8: '=2', 9: '=40', 10: '=254|>=254&<=300',
             11: '=40|=48', 12: '=1'}, IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_prefix(self):
        raw_hex = b'\x01\x18\x6e\x01\x01'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_DST_PFIX: '110.1.1.0/24'}, IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_packet_length(self):
        raw_hex = '\x0a\x01\xfe\x03\xfe\xd5\x01\x2c'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_PCK_LEN: '=254|>=254&<=300'}, IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_ip_protocol(self):
        raw_hex = '\x03\x01\x00\x01\x2f\x01\x58\x01\x01\x01\x02\x01\x59\x81\x67'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_IP_PROTO: '=0|=47|=88|=1|=2|=89|=103'}, IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_des_port(self):
        raw_hex = '\x05\x01\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_DST_PORT: '=80|=8080|=8081|=8082|=8083'},
                         IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_src_port(self):
        raw_hex = '\x06\x01\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_SRC_PORT: '=80|=8080|=8081|=8082|=8083'},
                         IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_icmp_type(self):
        raw_hex = '\x07\x01\x02\x01\x03\x01\x05\x81\x06'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_ICMP_TP: '=2|=3|=5|=6'}, IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_icmp_code(self):
        raw_hex = '\x08\x81\x02'
        self.assertEqual({bgp_cons.BGPNLRI_FSPEC_ICMP_CD: '=2'}, IPv4FlowSpec.parse(raw_hex))

    def test_construct_prefix(self):
        prefix_bin = b'\x18\xc0\x55\x02'
        prefix_str = '192.85.2.0/24'
        self.assertEqual(prefix_bin, IPv4FlowSpec.construct_prefix(prefix_str))

    def test_construct_operator_flag(self):
        flag = {
            'LEN': 1,
            'EQ': 1,
            'EOL': 1
        }
        self.assertEqual(0x81, IPv4FlowSpec.construct_operator_flag(flag))
        flag = {
            'EOL': 1,
            'LEN': 2,
            'AND': 1,
            'LT': 1,
            'EQ': 1
        }
        self.assertEqual(0xd5, IPv4FlowSpec.construct_operator_flag(flag))

    def test_construct_operators(self):
        operators = '=80|=8080|=8083'
        operators_bin = b'\x01\x50\x11\x1f\x90\x91\x1f\x93'
        self.assertEqual(operators_bin, IPv4FlowSpec.construct_operators(operators))

    def test_construct_nlri(self):
        raw_hex = '\x0f\x05\x01\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93'
        nlri_dict = {bgp_cons.BGPNLRI_FSPEC_DST_PORT: '=80|=8080|=8081|=8082|=8083'}
        self.assertEqual(raw_hex, IPv4FlowSpec.construct_nlri(nlri_dict))
        nlri_bin = b'\x0a\x01\x18\xc0\x55\x02\x02\x18\xc0\x55\x01'
        nlri_list = {1: '192.85.2.0/24', 2: '192.85.1.0/24'}
        self.assertEqual(nlri_bin, IPv4FlowSpec.construct_nlri(nlri_list))

if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yabgp.tlv import TLV
from yabgp.net import IPAddress
from ..linkstate import LinkState


@LinkState.register(_type=1028)
@LinkState.register(_type=1029)
class LocalRouterID(TLV):
    """
    local ipv4 or ipv6 router id
    """
    TYPE_STR = 'local_router_id'

    @classmethod
    def unpack(cls, data):

        router_id = IPAddress.unpack(data)
        return cls(value=router_id)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""IPv4 Flowspec NLRI
"""
import binascii
import math
import struct

import netaddr

from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri import NLRI


class IPv4FlowSpec(NLRI):
    """ipv4 flow nlri process
    """

    @classmethod
    def parse(cls, value):
        """
        parse IPv4 flowspec NLRI
        :param value:
        :return:
        """
        # +------------------------------+
        # |    length (0xnn or 0xfn nn)  |
        # +------------------------------+
        # |    NLRI value  (variable)    |
        # +------------------------------+
        nlri_dict = {}
        while value:
            offset = 0
            flowspec_type = ord(value[0:1])
            offset += 1
            # decode all kinds of flow spec
            if flowspec_type in [bgp_cons.BGPNLRI_FSPEC_DST_PFIX, bgp_cons.BGPNLRI_FSPEC_SRC_PFIX]:
                prefix, offset_tmp = cls.parse_prefix(value[offset:])
                offset += offset_tmp
                nlri_dict[flowspec_type] = prefix
                value = value[offset:]
            elif flowspec_type in [bgp_cons.BGPNLRI_FSPEC_IP_PROTO, bgp_cons.BGPNLRI_FSPEC_DST_PORT,
                                   bgp_cons.BGPNLRI_FSPEC_SRC_PORT, bgp_cons.BGPNLRI_FSPEC_ICMP_TP,
                                   bgp_cons.BGPNLRI_FSPEC_ICMP_CD, bgp_cons.BGPNLRI_FSPEC_DSCP,
                                   bgp_cons.BGPNLRI_FSPEC_PCK_LEN]:
                operator_list, offset = cls.parse_operators(value[offset:])
                nlri_dict[flowspec_type] = cls.operator_dict_to_str(operator_list)
                value = value[offset:]
            else:
                operator_list, offset = cls.parse_operators(value[offset:])
                nlri_dict[flowspec_type] = cls.operator_dict_to_str(operator_list)
                value = value[offset:]
        return nlri_dict

    @classmethod
    def construct(cls, value):
        nlri_hex = b''
        for nlri in value:
            nlri_hex += cls.construct_nlri(nlri)
        return nlri_hex

    @classmethod
    def construct_nlri(cls, data):
        """ Construct NLRI """
        # there may have many filters in each nlri
        data = dict([(int(l), r) for (l, r) in data.items()])
        nlri_tmp = b''
        for type_tmp in [bgp_cons.BGPNLRI_FSPEC_DST_PFIX, bgp_cons.BGPNLRI_FSPEC_SRC_PFIX]:
            if data.get(type_tmp):
                nlri_tmp += struct.pack('!B', type_tmp) + cls.construct_prefix(data[type_tmp])
        for type_tmp in [bgp_cons.BGPNLRI_FSPEC_IP_PROTO, bgp_cons.BGPNLRI_FSPEC_DST_PORT,
                         bgp_cons.BGPNLRI_FSPEC_SRC_PORT, bgp_cons.BGPNLRI_FSPEC_ICMP_TP,
                         bgp_cons.BGPNLRI_FSPEC_ICMP_CD, bgp_cons.BGPNLRI_FSPEC_PCK_LEN,
                         bgp_cons.BGPNLRI_FSPEC_DSCP, bgp_cons.BGPNLRI_FSPEC_PORT]:
            if not data.get(type_tmp):
                continue

            # translate from expression to binary
            nlri_tmp += struct.pack('!B', type_tmp) + cls.construct_operators(data[type_tmp])

        if len(nlri_tmp) >= 240:
            return struct.pack('!H', len(nlri_tmp)) + nlri_tmp
        elif nlri_tmp:
            return struct.pack('!B', len(nlri_tmp)) + nlri_tmp

    @staticmethod
    def parse_prefix(data):
        """
        Prefixes are encoded as in BGP UPDATE messages, a length in bits is followed by
        enough octets to contain the prefix information.

        Encoding: <prefix-length (1 octet), prefix>
        """
        prefix_len = ord(data[0:1])
        octet_len = int(math.ceil(prefix_len / 8.0))
        tmp = data[1:octet_len + 1]
        prefix_data = [ord(i) for i in tmp]
        prefix_data = prefix_data + list(str(0)) * 4
        prefix = "%s.%s.%s.%s" % (tuple(prefix_data[0:4])) + '/' + str(prefix_len)
        return prefix, octet_len + 1

    @classmethod
    def construct_prefix(cls, prefix):
        """
        construct a prefix string from '1.1.1.0/24' to '\x18\x01\x01\x01'
        """
        ip, masklen = prefix.split('/')
        ip_hex = netaddr.IPAddress(ip).packed
        masklen = int(masklen)
        if 16 < masklen <= 24:
            ip_hex = ip_hex[0:3]
        elif 8 < masklen <= 16:
            ip_hex = ip_hex[0:2]
        elif 0 < masklen <= 8:
            ip_hex = ip_hex[0:1]
        elif masklen == 0:
            ip_hex = ''
        return struct.pack('!B', masklen) + ip_hex

    @classmethod
    def parse_operators(cls, data):
        offset = 0
        parse_operator_list = []
        while data:
            operator = cls.parse_operator_flag(ord(data[0:1]))
            offset += 1
            operator_value = int(binascii.b2a_hex(data[1:1 + operator['LEN']]), 16)
            offset += operator['LEN']
            parse_operator_list.append([operator, operator_value])
            # the end of the list
            data = data[1 + operator['LEN']:]
            if operator['EOL']:
                break
        return parse_operator_list, offset + 1

    @staticmethod
    def parse_operator_flag(data):
        """
        The operator byte is encoded as:
        0    1   2   3   4  5   6   7
        +---+---+---+---+---+---+---+---+
        |EOL|AND|  LEN  |RES|LT |GT |EQ |
        +---+---+---+---+---+---+---+---+
        """
        bit_list = []
        for i in xrange(8):
            bit_list.append((data >> i) & 1)
        bit_list.reverse()
        result = {
            'EOL': bit_list[0],
            'AND': bit_list[1],
            'LEN': 1 << (bit_list[2] * 2 + bit_list[3]),
            'LT': bit_list[5],
            'GT': bit_list[6],
            'EQ': bit_list[7]
        }
        return result

    @staticmethod
    def construct_operator_flag(data):
        """construct operator flag from dict to binary
        """
        opt_dict = {
            'EOL': 0x80,
            'AND': 0x40,
            'LEN': {
                1: 0x00,
                2: 0x10,
                4: 0x20,
                6: 0x30
            },
            'RES': 0x00,
            'LT': 0x04,
            'GT': 0x02,
            'EQ': 0x01
        }
        b_data = 0x00
        for opt in opt_dict:
            if opt in data and opt != 'LEN':
                if data[opt] == 1:
                    b_data += opt_dict[opt]
            elif opt == 'LEN' and data[opt]:
                b_data += opt_dict['LEN'][data['LEN']]
        return b_data

    @staticmethod
    def operator_dict_to_str(data):
        """

        from
        [
            [
                {'AND': 0, 'GT': 0, 'LEN': 1, 'EOL': 0, 'LT': 0, 'EQ': 1},
                254
            ],
            [
                {'AND': 0, 'GT': 1, 'LEN': 1, 'EOL': 0, 'LT': 0, 'EQ': 1},
                254
            ],
            [
                {'AND': 1, 'GT': 0, 'LEN': 2, 'EOL': 1, 'LT': 1, 'EQ': 1},
                300
            ]
        ]
        to
        =254|>=254&<=300
        :param data: dict
        :return: string format
        """
        return_str = ''
        for item in data:
            operator_dict, value = item
            if operator_dict['AND']:
                return_str += '&'
            else:
                if return_str != '':
                    return_str += '|'
            if operator_dict['GT']:
                return_str += '>'
            if operator_dict['LT']:
                return_str += '<'
            if operator_dict['EQ']:
                return_str += '='
            return_str += str(value)
        return return_str

    @classmethod
    def construct_operators(cls, data):
        """
        from "=254|>=254&<=300" to binary data
        :param data:
        :return:
        """
        data_bin = b''
        data_list = data.split('|')
        eol = 0
        for i, data in enumerate(data_list):
            if i == len(data_list) - 1:
                eol = 1
            if '&' not in data:
                flag_dict = {'EOL': eol}
                if data[0] == '=':
                    off_set = 1
                    flag_dict['EQ'] = 1
                elif '>=' in data:
                    off_set = 2
                    flag_dict['EQ'] = 1
                    flag_dict['GT'] = 1
                elif '<=' in data:
                    off_set = 2
                    flag_dict['EQ'] = 1
                    flag_dict['LT'] = 1
                elif '>' in data:
                    off_set = 1
                    flag_dict['GT'] = 1
                elif '<' in data:
                    off_set = 1
                    # value_hex = str(bytearray.fromhex(hex(int(data[2:]))[2:]))
                    flag_dict['LT'] = 1
                hex_str = hex(int(data[off_set:]))[2:]
                if len(hex_str) % 2 == 1:
                    hex_str = '0' + hex_str
                value_hex = str(bytearray.fromhex(hex_str))
                flag_dict['LEN'] = len(value_hex)
                opt_flag_bin = cls.construct_operator_flag(flag_dict)
                data_bin += struct.pack('!B', opt_flag_bin)
                data_bin += value_hex

        return data_bin



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Attribute MP_REACH_NLRI
"""

import struct
import binascii

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.common import afn
from yabgp.common import safn
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri.ipv4_mpls_vpn import IPv4MPLSVPN
from yabgp.message.attribute.nlri.ipv6_mpls_vpn import IPv6MPLSVPN
from yabgp.message.attribute.nlri.ipv4_flowspec import IPv4FlowSpec
from yabgp.message.attribute.nlri.ipv4_srte import IPv4SRTE
from yabgp.message.attribute.nlri.ipv6_unicast import IPv6Unicast
from yabgp.message.attribute.nlri.labeled_unicast.ipv4 import IPv4LabeledUnicast
from yabgp.message.attribute.nlri.labeled_unicast.ipv6 import IPv6LabeledUnicast
from yabgp.message.attribute.nlri.evpn import EVPN
from yabgp.message.attribute.nlri.linkstate import BGPLS


class MpReachNLRI(Attribute):
    """
        MP_REACH_NLRI (type code 14) is used to carry the set of reachable
    destinations together with the next hop information to be used for
    forwarding to these destinations (RFC 4760 page 2).

        MP_REACH_NLRI coding information
        +---------------------------------------------------------+
        | Address Family Identifier (2 octets)                    |
        +---------------------------------------------------------+
        | Subsequent Address Family Identifier (1 octet)          |
        +---------------------------------------------------------+
        | Length of Next Hop Network Address (1 octet)            |
        +---------------------------------------------------------+
        | Network Address of Next Hop (variable)                  |
        +---------------------------------------------------------+
        | Reserved (1 octet)                                      |
        +---------------------------------------------------------+
        | Network Layer Reachability Information (variable)       |
        +---------------------------------------------------------+
    """

    ID = AttributeID.MP_REACH_NLRI
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.EXTENDED_LENGTH

    @classmethod
    def parse(cls, value):
        """parse
        """
        try:
            afi, safi, nexthop_length = struct.unpack('!HBB', value[0:4])
            nexthop_bin = value[4:4 + nexthop_length]
            nlri_bin = value[5 + nexthop_length:]
        except Exception:
            # error when lenght is wrong
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=str(value))

        #  Address Family IPv4
        if afi == afn.AFNUM_INET:
            if safi == safn.SAFNUM_LAB_VPNUNICAST:
                # MPLS VPN
                # parse nexthop
                rd_bin = nexthop_bin[0:8]
                rd_type = struct.unpack('!H', rd_bin[0:2])[0]
                rd_value_bin = rd_bin[2:]
                if rd_type == 0:
                    asn, an = struct.unpack('!HI', rd_value_bin)
                    ipv4 = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin[8:]), 16)))
                    nexthop = {'rd': '%s:%s' % (asn, an), 'str': ipv4}
                else:
                    nexthop = binascii.b2a_hex((nexthop_bin[8:]))
                # parse nlri
                nlri = IPv4MPLSVPN.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            elif safi == safn.SAFNUM_MPLS_LABEL:
                if nexthop_bin:
                    nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin), 16)))
                else:
                    nexthop = ''
                nlri = IPv4LabeledUnicast.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            elif safi == safn.SAFNUM_FSPEC_RULE:
                # if nlri length is greater than 240 bytes, it is encoded over 2 bytes
                nlri_list = []
                while nlri_bin:
                    length = ord(nlri_bin[0])
                    if length >> 4 == 0xf and len(nlri_bin) > 2:
                        length = struct.unpack('!H', nlri_bin[:2])[0]
                        nlri_tmp = nlri_bin[2: length + 2]
                        nlri_bin = nlri_bin[length + 2:]
                    else:
                        nlri_tmp = nlri_bin[1: length + 1]
                        nlri_bin = nlri_bin[length + 1:]
                    nlri = IPv4FlowSpec.parse(nlri_tmp)
                    if nlri:
                        nlri_list.append(nlri)
                if nexthop_bin:
                    nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin), 16)))
                else:
                    nexthop = ''
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri_list)
            else:
                nlri = binascii.b2a_hex(nlri_bin)

        # #  Address Family IPv6
        elif afi == afn.AFNUM_INET6:
            # IPv6 unicast
            if safi == safn.SAFNUM_UNICAST:
                # decode nexthop
                # RFC 2545
                # The value of the Length of Next Hop Network Address field on a
                # MP_REACH_NLRI attribute shall be set to 16, when only a global
                # address is present, or 32 if a link-local address is also included in
                # the Next Hop field.
                #
                # The link-local address shall be included in the Next Hop field if and
                # only if the BGP speaker shares a common subnet with the entity
                # identified by the global IPv6 address carried in the Network Address
                # of Next Hop field and the peer the route is being advertised to.
                nexthop_addrlen = 16
                has_link_local = False
                nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin[:nexthop_addrlen]), 16)))
                if len(nexthop_bin) == 2 * nexthop_addrlen:
                    # has link local address
                    has_link_local = True
                    linklocal_nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin[nexthop_addrlen:]), 16)))
                nlri = IPv6Unicast.parse(nlri_bin)
                if has_link_local:
                    return dict(afi_safi=(afi, safi), nexthop=nexthop, linklocal_nexthop=linklocal_nexthop, nlri=nlri)
                else:
                    return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            elif safi == safn.SAFNUM_LAB_VPNUNICAST:
                # IPv6 MPLS VPN
                # parse nexthop
                rd_bin = nexthop_bin[0:8]
                rd_type = struct.unpack('!H', rd_bin[0:2])[0]
                rd_value_bin = rd_bin[2:]
                if rd_type == 0:
                    asn, an = struct.unpack('!HI', rd_value_bin)
                    ipv6 = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin[8:]), 16)))
                    nexthop = {'rd': '%s:%s' % (asn, an), 'str': ipv6}
                # TODO(xiaoquwl) for other RD type decoding
                else:
                    nexthop = binascii.b2a_hex(nexthop_bin[8:])
                # parse nlri
                nlri = IPv6MPLSVPN.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            elif safi == safn.SAFNUM_MPLS_LABEL:
                if nexthop_bin:
                    nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin), 16)))
                else:
                    nexthop = ''
                nlri = IPv6LabeledUnicast.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            else:
                return dict(afi_safi=(afi, safi), nexthop=nexthop_bin, nlri=nlri_bin)

        # for l2vpn
        elif afi == afn.AFNUM_L2VPN:
            if safi == safn.SAFNUM_EVPN:
                nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin), 16)))
                nlri = EVPN.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            else:
                nlri = binascii.b2a_hex(nlri_bin)

        # BGP LS
        elif afi == afn.AFNUM_BGPLS:
            if safi == safn.SAFNUM_BGPLS:
                nexthop = str(netaddr.IPAddress(int(binascii.b2a_hex(nexthop_bin), 16)))
                nlri = BGPLS.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), nexthop=nexthop, nlri=nlri)
            else:
                pass
        else:
            nlri = binascii.b2a_hex(nlri_bin)

        return dict(afi_safi=(afi, safi), nexthop=binascii.b2a_hex(nexthop_bin), nlri=binascii.b2a_hex(nlri_bin))

    @classmethod
    def construct_mpls_vpn_nexthop(cls, nexthop):
        """
        construct nexthop to bin
        :param nexthop:
        :return:
        """
        asn, an = map(int, nexthop['rd'].split(':'))
        return struct.pack('!H', 0) + struct.pack('!HI', asn, an) + netaddr.IPAddress(nexthop['str']).packed

    @classmethod
    def construct(cls, value):

        """Construct a attribute

        :param value: python dictionary
        {'afi_safi': (1,128),
         'nexthop': {},
         'nlri': []
        """
        afi, safi = value['afi_safi']

        # for ipv4 unicast
        if afi == afn.AFNUM_INET:
            if safi == safn.SAFNUM_LAB_VPNUNICAST:  # MPLS VPN
                nexthop_hex = cls.construct_mpls_vpn_nexthop(value['nexthop'])
                nlri_hex = IPv4MPLSVPN.construct(value['nlri'])
                attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) +\
                    struct.pack('!B', len(nexthop_hex)) + nexthop_hex + b'\x00' + nlri_hex
                return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                    + struct.pack('!H', len(attr_value)) + attr_value
            elif safi == safn.SAFNUM_FSPEC_RULE:  # BGP Flow spec
                try:
                    try:
                        nexthop = netaddr.IPAddress(value['nexthop']).packed
                    except netaddr.core.AddrFormatError:
                        nexthop = ''
                    nlri_hex = b''
                    nlri_hex += IPv4FlowSpec.construct(value=value['nlri'])
                    if nlri_hex:
                        attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + \
                            struct.pack('!B', len(nexthop)) + nexthop + b'\x00' + nlri_hex
                        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                            + struct.pack('!H', len(attr_value)) + attr_value
                except Exception as e:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes: %s' % e,
                        data=value
                    )
            elif safi == safn.SAFNUM_SRTE:  # BGP SR TE Policy
                try:
                    try:
                        nexthop = netaddr.IPAddress(value['nexthop']).packed
                    except netaddr.core.AddrFormatError:
                        nexthop = ''
                    nlri_hex = b''
                    nlri_hex += IPv4SRTE.construct(data=value['nlri'])
                    if nlri_hex:
                        attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + \
                            struct.pack('!B', len(nexthop)) + nexthop + b'\x00' + nlri_hex
                        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                            + struct.pack('!H', len(attr_value)) + attr_value
                except Exception as e:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes: %s' % e,
                        data=value
                    )
            elif safi == safn.SAFNUM_MPLS_LABEL:
                try:
                    try:
                        nexthop = netaddr.IPAddress(value['nexthop']).packed
                    except netaddr.core.AddrFormatError:
                        nexthop = ''
                    nlri_hex = b''
                    nlri_hex += IPv4LabeledUnicast.construct(value['nlri'])
                    if nlri_hex:
                        attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + \
                            struct.pack('!B', len(nexthop)) + nexthop + b'\x00' + nlri_hex
                        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                            + struct.pack('!H', len(attr_value)) + attr_value
                except Exception as e:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes: %s' % e,
                        data=value
                    )
            else:
                raise excep.ConstructAttributeFailed(
                    reason='unsupport this sub address family',
                    data=value)

        # ipv6 unicast
        elif afi == afn.AFNUM_INET6:

            if safi == safn.SAFNUM_LAB_VPNUNICAST:  # MPLS VPN
                nexthop_hex = cls.construct_mpls_vpn_nexthop(value['nexthop'])
                nlri_hex = IPv6MPLSVPN.construct(value['nlri'])
                attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) +\
                    struct.pack('!B', len(nexthop_hex)) + nexthop_hex + b'\x00' + nlri_hex
                return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                    + struct.pack('!H', len(attr_value)) + attr_value

            elif safi == safn.SAFNUM_UNICAST:
                nexthop_len = 16
                nexthop_bin = netaddr.IPAddress(value['nexthop']).packed
                if value.get('linklocal_nexthop'):
                    nexthop_len *= 2
                    nexthop_bin += netaddr.IPAddress(value['linklocal_nexthop']).packed

                nlri_bin = IPv6Unicast.construct(nlri_list=value['nlri'])

                attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + struct.pack('!B', nexthop_len) + \
                    nexthop_bin + b'\x00' + nlri_bin
                return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID)\
                    + struct.pack('!H', len(attr_value)) + attr_value
            elif safi == safn.SAFNUM_MPLS_LABEL:
                try:
                    try:
                        nexthop = netaddr.IPAddress(value['nexthop']).packed
                    except netaddr.core.AddrFormatError:
                        nexthop = ''
                    nlri_hex = b''
                    nlri_hex += IPv6LabeledUnicast.construct(value['nlri'])
                    if nlri_hex:
                        attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + \
                            struct.pack('!B', len(nexthop)) + nexthop + b'\x00' + nlri_hex
                        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                            + struct.pack('!H', len(attr_value)) + attr_value
                except Exception as e:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes: %s' % e,
                        data=value
                    )
        # for l2vpn
        elif afi == afn.AFNUM_L2VPN:
            if safi == safn.SAFNUM_EVPN:
                nexthop_bin = netaddr.IPAddress(value['nexthop']).packed
                nlri_bin = EVPN.construct(nlri_list=value['nlri'])
                attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + struct.pack('!B', len(nexthop_bin)) + \
                    nexthop_bin + b'\x00' + nlri_bin
                return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID)\
                    + struct.pack('!H', len(attr_value)) + attr_value
        else:
            raise excep.ConstructAttributeFailed(
                reason='unsupport this sub address family',
                data=value)



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yabgp.message.attribute.nlri.mpls_vpn import MPLSVPN
from yabgp.common.afn import AFNUM_INET6
from yabgp.common.safn import SAFNUM_LAB_VPNUNICAST


class IPv6MPLSVPN(MPLSVPN):
    """
    IPv6 MPLS VPN NLRI
    """
    AFI = AFNUM_INET6
    SAFI = SAFNUM_LAB_VPNUNICAST

    @classmethod
    def parse(cls, value, iswithdraw=False):
        return super(IPv6MPLSVPN, cls).parse(value, iswithdraw=iswithdraw)

    @classmethod
    def construct(cls, value, iswithdraw=False):
        return super(IPv6MPLSVPN, cls).construct(value, iswithdraw=iswithdraw)



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii

import netaddr

from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/draft-gredler-idr-bgp-ls-segment-routing-ext-03#section-2.2.2
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |              Type             |            Length             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     Flags     |     Weight    |            Reserved           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |             OSPF Neighbor ID / IS-IS System-ID                |
# +                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                    SID/Label/Index (variable)                 |
# +---------------------------------------------------------------+

#   Flags: 1 octet field of following flags:
#       0 1 2 3 4 5 6 7
#      +-+-+-+-+-+-+-+-+
#      |F|B|V|L|S|P|   |
#      +-+-+-+-+-+-+-+-+


@LinkState.register()
class LanAdjSegID(TLV):
    """
    Adjacency Segment id
    """
    TYPE = 1100
    TYPE_STR = 'lan_adj_sid'

    @classmethod
    def unpack(cls, data, pro_id):
        flags = ord(data[0:1])
        weight = ord(data[1:2])
        flag = {}
        if pro_id in [1, 2]:
            flag['F'] = flags >> 7
            flag['B'] = (flags << 1) % 256 >> 7
            flag['V'] = (flags << 2) % 256 >> 7
            flag['L'] = (flags << 3) % 256 >> 7
            flag['S'] = (flags << 4) % 256 >> 7
            flag['P'] = (flags << 5) % 256 >> 7
            tmp = binascii.b2a_hex(data[4:10])
            chunks, chunk_size = len(tmp), len(tmp)/3
            nei_or_sys_id = '.'.join([tmp[i:i+chunk_size] for i in range(0, chunks, chunk_size)])
            sid_index_label = int(binascii.b2a_hex(data[10:]), 16)
        else:  # 3, 6
            flag['B'] = flags >> 7
            flag['V'] = (flags << 1) % 256 >> 7
            flag['L'] = (flags << 2) % 256 >> 7
            flag['G'] = (flags << 3) % 256 >> 7
            flag['P'] = (flags << 4) % 256 >> 7
            nei_or_sys_id = str(netaddr.IPAddress(int(binascii.b2a_hex(data[4:8]), 16)))
            sid_index_label = int(binascii.b2a_hex(data[8:]), 16)
        return cls(value={
            "flags": flag,
            "weight": weight,
            "nei_or_sys_id": nei_or_sys_id,
            "value": sid_index_label
        })



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii

from yabgp.message.attribute import Attribute, AttributeFlag, AttributeID


class LinkState(Attribute):
    """BGP linksate attribute
    """
    ID = AttributeID.LINKSTATE
    FLAG = AttributeFlag.OPTIONAL

    registered_tlvs = dict()

    def __init__(self, value, hex_value=None):
        self.value = value
        self.hex_value = hex_value

    @classmethod
    def register(cls, _type=None):
        """register tlvs
        """
        def decorator(klass):
            """decorator
            """
            _id = klass.TYPE if _type is None else _type
            if _id in cls.registered_tlvs:
                raise RuntimeError('duplicated attribute type')
            cls.registered_tlvs[_id] = klass
            return klass
        return decorator

    def dict(self):
        """return dict
        """
        return {self.ID: self.value}

    @classmethod
    def unpack(cls, data, bgpls_pro_id=None):
        """unpack binary data
        """
        tlvs = []
        while data:
            type_code, length = struct.unpack('!HH', data[:4])
            value = data[4: 4+length]
            if type_code in [1099, 1100] and type_code in cls.registered_tlvs:
                tlvs.append(cls.registered_tlvs[type_code].unpack(value, bgpls_pro_id).dict())
            elif type_code in cls.registered_tlvs:
                tlvs.append(cls.registered_tlvs[type_code].unpack(value).dict())
            else:
                tlvs.append(
                    {
                        'type': type_code,
                        'value': str(binascii.b2a_hex(value))

                    }
                )
            data = data[4+length:]

        return cls(value=tlvs)



# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.message.attribute.nlri.labeled_unicast.ipv4 import IPv4LabeledUnicast


class TestIPv4LabeledUnicast(unittest.TestCase):

    def test_construct(self):
        nlri_list = [
            {'prefix': '34.1.41.0/24', 'label': [321]},
            {'prefix': '34.1.42.0/24', 'label': [322]}
        ]
        nlri_hex = b'\x30\x00\x14\x11\x22\x01\x29\x30\x00\x14\x21\x22\x01\x2a'
        self.assertEqual(nlri_hex, IPv4LabeledUnicast.construct(nlri_list))

    def test_construct_with_multi_label(self):
        nlri_list = [
            {'prefix': '34.1.41.0/24', 'label': [321, 322]},
            {'prefix': '34.1.42.0/24', 'label': [321, 322]}
        ]
        nlri_hex = b'\x48\x00\x14\x10\x00\x14\x21\x22\x01\x29\x48\x00\x14\x10\x00\x14\x21\x22\x01\x2a'
        self.assertEqual(nlri_hex, IPv4LabeledUnicast.construct(nlri_list))

    def test_parse(self):
        nlri_list = [
            {'prefix': '34.1.41.0/24', 'label': [321]},
            {'prefix': '34.1.42.0/24', 'label': [322]}
        ]
        nlri_hex = b'\x30\x00\x14\x11\x22\x01\x29\x30\x00\x14\x21\x22\x01\x2a'
        self.assertEqual(nlri_list, IPv4LabeledUnicast.parse(nlri_hex))

    def test_parse_with_multi_label(self):
        nlri_list = [
            {'prefix': '34.1.41.0/24', 'label': [321, 322]},
            {'prefix': '34.1.42.0/24', 'label': [321, 322]}
        ]
        nlri_hex = b'\x48\x00\x14\x10\x00\x14\x21\x22\x01\x29\x48\x00\x14\x10\x00\x14\x21\x22\x01\x2a'
        self.assertEqual(nlri_list, IPv4LabeledUnicast.parse(nlri_hex))


if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/draft-ietf-idr-bgp-ls-segment-routing-ext-03#section-2.1.1
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |               Type            |            Length             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                      SID/Label (variable)                     |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# If length is set to 3, then the 20 rightmost bits represent a label.
# If length is set to 4, then the value represents a 32 bit SID.


@LinkState.register()
class SIDorLabel(TLV):

    TYPE = 1161  # https://tools.ietf.org/html/draft-ietf-idr-bgp-ls-segment-routing-ext-03#section-2.1.1
    TYPE_STR = "sid_or_label"

    @classmethod
    def unpack(cls, value):
        """
        """
        length = len(value)
        if length == 3:
            return cls(value={"type": "sid", "value": (struct.unpack('!I', "\x00" + value)[0] << 12) >> 12})
        elif length == 4:
            return cls(value={"type": "label", "value": struct.unpack('!I', value)[0]})



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Blueprint for version 1 of API
"""
import logging
import time

from flask.ext.httpauth import HTTPBasicAuth
from flask import Blueprint
import flask
from oslo_config import cfg

from yabgp.api import utils as api_utils
from yabgp.common import constants as bgp_cons

LOG = logging.getLogger(__name__)
blueprint = Blueprint('v1', __name__)
auth = HTTPBasicAuth()


@auth.get_password
def get_pw(username):
    if username == cfg.CONF.rest.username:
        return cfg.CONF.rest.password
    return None


@blueprint.route('/')
@api_utils.log_request
def root():
    """
    v1 api root. Get the api status.
    """
    intro = {
        "status": "stable",
        "updated": "2015-01-22T00:00:00Z",
        "version": "v1"}
    cfg.CONF.keep_alive.last_time = time.time()
    return flask.jsonify(intro)


@blueprint.route('/peer/<peer_ip>/state')
@auth.login_required
@api_utils.log_request
def peer(peer_ip):
    """
    Get one peer's running information, include basic configurations and fsm state.
    """
    return flask.jsonify(api_utils.get_peer_conf_and_state(peer_ip))


@blueprint.route('/peer/<peer_ip>/statistic')
@auth.login_required
@api_utils.log_request
def get_peer_statistic(peer_ip):
    """
    Get one peer's message statistic, include sending and receiving.
    """
    return flask.jsonify(api_utils.get_peer_msg_statistic(peer_ip))


@blueprint.route('/peer/<peer_ip>/send/route-refresh', methods=['POST'])
@auth.login_required
@api_utils.log_request
@api_utils.makesure_peer_establish
def send_route_refresh(peer_ip):
    """
    Try to send BGP Route Refresh message to a peer
    """
    LOG.debug('Try to send route refresh to peer %s', peer_ip)
    json_request = flask.request.get_json()
    if 'afi' in json_request and 'safi' in json_request:
        if 'res' not in json_request:
            res = 0
        else:
            res = json_request['res']
        result = api_utils.send_route_refresh(
            peer_ip=peer_ip, afi=json_request['afi'], safi=json_request['safi'], res=res)
        return flask.jsonify(result)
    return flask.jsonify({
        'status': False,
        'code': 'please check your post data'
    })


@blueprint.route('/peer/<peer_ip>/send/update', methods=['POST'])
@auth.login_required
@api_utils.log_request
@api_utils.makesure_peer_establish
def send_update_message(peer_ip):
    """
    Try to send BGP update message to the peer.
    Both update nlri and withdraw nlri treated as Update.
    """
    LOG.debug('Try to send update message to peer %s', peer_ip)
    json_request = flask.request.get_json()
    attr = json_request.get('attr')
    nlri = json_request.get('nlri')
    withdraw = json_request.get('withdraw')
    if attr:
        attr = {int(k): v for k, v in attr.items()}
        res = api_utils.get_peer_conf_and_state(peer_ip)
        if 5 not in attr and res['peer']['remote_as'] == res['peer']['local_as']:
            # default local preference
            attr[5] = 100
        if 16 in attr:
            # extended community recombine
            ext_community = []
            for ext_com in attr[16]:
                key, value = ext_com.split(':', 1)
                if key.strip().lower() == 'route-target':
                    values = value.strip().split(',')
                    for vau in values:
                        if '.' in vau.strip().split(':')[0]:
                            ext_community.append([258, vau.strip()])
                        else:
                            nums = vau.strip().split(':', 1)
                            if int(nums[1].strip()) <= 65535:
                                ext_community.append([2, vau.strip()])
                            else:
                                if res['peer']['capability']['remote']:
                                    four_bytes_as = res['peer']['capability']['remote']['four_bytes_as']
                                else:
                                    return flask.jsonify({
                                        'status': False,
                                        'code': 'please check peer state'
                                    })
                                if int(nums[1].strip()) > 65535 and four_bytes_as:
                                    ext_community.append([514, vau.strip()])
                                elif not four_bytes_as and int(nums[0].strip()) > 65535:
                                    return flask.jsonify({
                                        'status': False,
                                        'code': 'peer not support as num of greater than 65535'
                                    })
                elif key.strip().lower() == 'route-origin':
                    values = value.strip().split(',')
                    for vau in values:
                        if '.' in vau.strip().split(':')[0]:
                            ext_community.append([259, vau.strip()])
                        else:
                            if res['peer']['capability']['remote']:
                                four_bytes_as = res['peer']['capability']['remote']['four_bytes_as']
                            else:
                                return flask.jsonify({
                                    'status': False,
                                    'code': 'please check peer state'
                                })
                            nums = vau.strip().split(':', 1)
                            if int(nums[1].strip()) <= 65535 and four_bytes_as:
                                ext_community.append([515, vau.strip()])
                            elif not four_bytes_as and int(nums[0].strip()) > 65535:
                                return flask.jsonify({
                                    'status': False,
                                    'code': 'peer not support as num of greater than 65535'
                                })
                            else:
                                ext_community.append([3, vau.strip()])
                elif key.strip().lower() == 'redirect-nexthop':
                    ext_community.append([2048, '0.0.0.0', int(value.strip())])
                elif bgp_cons.BGP_EXT_COM_DICT_1.get(key.strip().lower()):
                    key_num = bgp_cons.BGP_EXT_COM_DICT_1.get(key.strip().lower())
                    values = value.strip().split(':', 1)
                    ext_community.append([key_num, int(values[0]), int(values[1])])
                else:
                    key_num = bgp_cons.BGP_EXT_COM_DICT.get(key.strip().lower())
                    if key_num:
                        values = value.strip().split(',')
                        for vau in values:
                            if key_num == 32777:
                                ext_community.append([key_num, int(vau.strip())])
                            else:
                                ext_community.append([key_num, vau.strip()])
                    else:
                        return flask.jsonify({
                            'status': False,
                            'code': 'unexpected extended community "%s", please check your post data' % key
                        })
            attr[16] = ext_community
    if (attr and nlri) or withdraw:
        return flask.jsonify(api_utils.send_update(peer_ip, attr, nlri, withdraw))
    elif 14 in attr or 15 in attr:
        return flask.jsonify(api_utils.send_update(peer_ip, attr, nlri, withdraw))

    else:
        return flask.jsonify({
            'status': False,
            'code': 'please check your post data'
        })



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test PMSI Tunnel attribute
"""

import unittest

from yabgp.message.attribute.pmsitunnel import PMSITunnel


class TestPMSITunnel(unittest.TestCase):

    def test_parse(self):
        self.maxDiff = None
        hex_value = b'\x00\x06\x00\x27\x10\x04\x04\x04\x04'
        value_hoped = {'mpls_label': [625], 'tunnel_id': '4.4.4.4', 'tunnel_type': 6, 'leaf_info_required': 0}
        self.assertEqual(value_hoped, PMSITunnel.parse(hex_value))

    def test_construct(self):
        self.maxDiff = None
        hex_value = b'\x00\x06\x00\x27\x10\x04\x04\x04\x04'
        value_hoped = {'mpls_label': [625], 'tunnel_id': '4.4.4.4', 'tunnel_type': 6, 'leaf_info_required': 0}
        self.assertEqual(hex_value, PMSITunnel.construct(value_hoped)[3:])

if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii

import netaddr

from yabgp.message.attribute.nlri import NLRI
from yabgp.common.afn import AFNUM_INET6, AFNUM_INET
from yabgp.common.safn import SAFNUM_MPLS_LABEL


class LabeledUnicast(NLRI):

    """
    +---------------------------+
    |   Length (1 octet)        |
    +---------------------------+
    |   Label (3 octets)        |
    +---------------------------+
    .............................
    +---------------------------+
    |   Prefix (variable)       |
    +---------------------------+
    """

    SAFI = SAFNUM_MPLS_LABEL

    @classmethod
    def parse(cls, nlri_data):
        nlri_list = []
        while nlri_data:
            nlri_bit_len = ord(nlri_data[0])
            # get the byte lenght of label stack + prefix
            if nlri_bit_len % 8 == 0:
                nlri_byte_len = nlri_bit_len / 8
            else:
                nlri_byte_len = nlri_bit_len / 8 + 1

            offset = nlri_byte_len + 1
            label = cls.parse_mpls_label_stack(nlri_data[1:])
            label_byte_len = len(label) * 3
            prefix_byte_len = nlri_byte_len - label_byte_len
            prefix_mask = nlri_bit_len - label_byte_len * 8
            if cls.AFI == AFNUM_INET:
                prefix_hex = nlri_data[offset - prefix_byte_len: offset] + (4 - prefix_byte_len) * '\x00'
                prefix_data = [ord(i) for i in prefix_hex]
                prefix_data = prefix_data
                prefix = "%s.%s.%s.%s" % (tuple(prefix_data[0:4])) + '/' + str(prefix_mask)
            elif cls.AFI == AFNUM_INET6:  # ipv6
                prefix_hex = nlri_data[offset - prefix_byte_len: offset]
                prefix = str(netaddr.IPAddress(int(binascii.b2a_hex(prefix_hex), 16))) + '/' + str(prefix_mask)
            nlri_data = nlri_data[offset:]
            nlri_list.append({'prefix': prefix, 'label': label})
        return nlri_list

    @classmethod
    def construct(cls, nlri_list):
        nlri_list_hex = b''
        for nlri in nlri_list:
            label_hex = cls.construct_mpls_label_stack(nlri['label'])
            if cls.AFI == AFNUM_INET:
                prefixstr, prefixlen = nlri['prefix'].split('/')
                prefixlen = int(prefixlen)
                prefix_hex = cls.construct_prefix_v4(prefixlen, prefixstr)
            elif cls.AFI == AFNUM_INET6:
                prefixlen = int(nlri['prefix'].split('/')[1])
                prefix_hex = cls.construct_prefix_v6(nlri['prefix'])
            nlri_list_hex += struct.pack('!B', 8 * len(label_hex) + prefixlen) + label_hex + prefix_hex
        return nlri_list_hex



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/rfc7752#section-3.3.3.1
#  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |              Type             |             Length            |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |D|N|L|P| Resvd.|
# +-+-+-+-+-+-+-+-+

# +----------+---------------------------+-----------+
# |   Bit    | Description               | Reference |
# +----------+---------------------------+-----------+
# |   'D'    | IS-IS Up/Down Bit         | [RFC5305] |
# |   'N'    | OSPF "no unicast" Bit     | [RFC5340] |
# |   'L'    | OSPF "local address" Bit  | [RFC5340] |
# |   'P'    | OSPF "propagate NSSA" Bit | [RFC5340] |
# | Reserved | Reserved for future use.  |           |
# +----------+---------------------------+-----------+


@LinkState.register()
class IGPFlags(TLV):

    TYPE = 1152  # https://tools.ietf.org/html/rfc7752#section-3.3.3.1
    TYPE_STR = "igp_flags"

    @classmethod
    def unpack(cls, value):
        """
        """
        value = ord(value[0:1])
        D = value >> 7
        N = (value << 1) % 256 >> 7
        L = (value << 2) % 256 >> 7
        P = (value << 3) % 256 >> 7
        return cls(value={'D': D, 'N': N, 'L': L, 'P': P})



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""IPv4 SR TE Policy NLRI
"""
import struct
import netaddr

from yabgp.message.attribute.nlri import NLRI


class IPv4SRTE(NLRI):
    """
    """
    # +------------------+
    # |  NLRI Length     |    1 octet
    # +------------------+
    # |  Distinguisher   |    4 octets
    # +------------------+
    # |  Policy Color    |    4 octets
    # +------------------+
    # |  Endpoint        |    4 or 16 octets
    # +------------------+
    @classmethod
    def construct(cls, data):
        """ Construct NLRI """
        nlri_tmp = b'' + struct.pack('!I', data['distinguisher']) + \
            struct.pack('!I', data['color']) + \
            netaddr.IPAddress(data['endpoint']).packed
        return struct.pack('!B', len(nlri_tmp) * 8) + nlri_tmp



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" All BGP constant values """

# some handy things to know
BGP_MAX_PACKET_SIZE = 4096
BGP_MARKER_SIZE = 16  # size of BGP marker
BGP_HEADER_SIZE = 19  # size of BGP header, including marker
BGP_MIN_OPEN_MSG_SIZE = 29
BGP_MIN_UPDATE_MSG_SIZE = 23
BGP_MIN_NOTIFICATION_MSG_SIZE = 21
BGP_MIN_KEEPALVE_MSG_SIZE = BGP_HEADER_SIZE
BGP_TCP_PORT = 179
BGP_ROUTE_DISTINGUISHER_SIZE = 8

# BGP message types
BGP_OPEN = 1
BGP_UPDATE = 2
BGP_NOTIFICATION = 3
BGP_KEEPALIVE = 4
BGP_ROUTE_REFRESH = 5
BGP_CAPABILITY = 6
BGP_ROUTE_REFRESH_CISCO = 0x80

BGP_SIZE_OF_PATH_ATTRIBUTE = 2

# attribute flags, from RFC1771
BGP_ATTR_FLAG_OPTIONAL = 0x80
BGP_ATTR_FLAG_TRANSITIVE = 0x40
BGP_ATTR_FLAG_PARTIAL = 0x20
BGP_ATTR_FLAG_EXTENDED_LENGTH = 0x10


# SSA flags
BGP_SSA_TRANSITIVE = 0x8000
BGP_SSA_TYPE = 0x7FFF

# SSA Types
BGP_SSA_L2TPv3 = 1
BGP_SSA_mGRE = 2
BGP_SSA_IPSec = 3
BGP_SSA_MPLS = 4
BGP_SSA_L2TPv3_IN_IPSec = 5
BGP_SSA_mGRE_IN_IPSec = 6

# AS_PATH segment types
AS_SET = 1  # RFC1771
AS_SEQUENCE = 2  # RFC1771
AS_CONFED_SET = 4  # RFC1965 has the wrong values, corrected in
AS_CONFED_SEQUENCE = 3  # draft-ietf-idr-bgp-confed-rfc1965bis-01.txt

# OPEN message Optional Parameter types
BGP_OPTION_AUTHENTICATION = 1  # RFC1771
BGP_OPTION_CAPABILITY = 2  # RFC2842

# attribute types
BGPTYPE_ORIGIN = 1  # RFC1771
BGPTYPE_AS_PATH = 2  # RFC1771
BGPTYPE_NEXT_HOP = 3  # RFC1771
BGPTYPE_MULTI_EXIT_DISC = 4  # RFC1771
BGPTYPE_LOCAL_PREF = 5  # RFC1771
BGPTYPE_ATOMIC_AGGREGATE = 6  # RFC1771
BGPTYPE_AGGREGATOR = 7  # RFC1771
BGPTYPE_COMMUNITIES = 8  # RFC1997
BGPTYPE_ORIGINATOR_ID = 9  # RFC2796
BGPTYPE_CLUSTER_LIST = 10  # RFC2796
BGPTYPE_DPA = 11  # work in progress
BGPTYPE_ADVERTISER = 12  # RFC1863
BGPTYPE_RCID_PATH = 13  # RFC1863
BGPTYPE_MP_REACH_NLRI = 14  # RFC2858
BGPTYPE_MP_UNREACH_NLRI = 15  # RFC2858
BGPTYPE_EXTENDED_COMMUNITY = 16  # Draft Ramachandra
BGPTYPE_NEW_AS_PATH = 17  # draft-ietf-idr-as4bytes
BGPTYPE_NEW_AGGREGATOR = 18  # draft-ietf-idr-as4bytes
BGPTYPE_SAFI_SPECIFIC_ATTR = 19  # draft-kapoor-nalawade-idr-bgp-ssa-00.txt
BGPTYPE_PMSI_TUNNEL = 22  # RFC 6514
BGPTYPE_TUNNEL_ENCAPS_ATTR = 23  # RFC5512
BGPTYPE_LINK_STATE = 29
BGPTYPE_ATTRIBUTE_SET = 128

# BGP Tunnel Encapsulation Attribute Tunnel Types
BGP_TUNNEL_ENCAPS_RESERVED = 0
BGP_TUNNEL_ENCAPS_L2TPV3_OVER_IP = 1
BGP_TUNNEL_ENCAPS_GRE = 2
BGP_TUNNEL_ENCAPS_TRANSMIT_TUNNEL_ENDPOINT = 3
BGP_TUNNEL_ENCAPS_IPSEC_TUNNEL_MODE = 4
BGP_TUNNEL_ENCAPS_IP_IN_IP_TUNNEL_WITH_IPSEC = 5
BGP_TUNNEL_ENCAPS_MPLS_IN_IP_TUNNEL_WITH_IPSEC = 6
BGP_TUNNEL_ENCAPS_IP_IN_IP = 7
BGP_TUNNEL_ENCAPS_VXLAN = 8
BGP_TUNNEL_ENCAPS_NVGRE = 9
BGP_TUNNEL_ENCAPS_MPLS = 10
BGP_TUNNEL_ENCAPS_MPLS_IN_GRE = 11
BGP_TUNNEL_ENCAPS_VXLAN_GRE = 12
BGP_TUNNEL_ENCAPS_MPLS_IN_UDP = 13
BGP_TUNNEL_ENCAPS_IPV6_TUNNEL = 14
BGP_TUNNEL_ENCAPS_SR_TE_POLICY_TYPE = 15
BGP_TUNNEL_ENCAPS_BARE = 16

# Segment Sub-TLV type
BGP_SRTE_SEGMENT_SUBTLV_MPLS = 1
BGP_SRTE_SEGMENT_SUBTLV_IPV6 = 2
BGP_SRTE_SEGMENT_SUBTLV_IPV4_SID = 3
BGP_SRTE_SEGMENT_SUBTLV_IPV6_SID = 4
BGP_SRTE_SEGMENT_SUBTLV_IPV4_INDEX_SID = 5
BGP_SRTE_SEGMENT_SUBTLV_IPV4_ADDR_SID = 6
BGP_SRTE_SEGMENT_SUBTLV_IPV6_INDEX_SID = 7
BGP_SRTE_SEGMENT_SUBTLV_IPV6_ADDR_SID = 8

#  VPN Route Target  #
BGP_EXT_COM_RT_0 = 0x0002  # Route Target,Format AS(2bytes):AN(4bytes)
BGP_EXT_COM_RT_1 = 0x0102  # Route Target,Format IPv4 address(4bytes):AN(2bytes)
BGP_EXT_COM_RT_2 = 0x0202  # Route Target,Format AS(4bytes):AN(2bytes)

# Route Origin (SOO site of Origin)
BGP_EXT_COM_RO_0 = 0x0003  # Route Origin,Format AS(2bytes):AN(4bytes)
BGP_EXT_COM_RO_1 = 0x0103  # Route Origin,Format IP address:AN(2bytes)
BGP_EXT_COM_RO_2 = 0x0203  # Route Origin,Format AS(4bytes):AN(2bytes)

# BGP Flow Spec
BGP_EXT_REDIRECT_NH = 0x0800  # redirect to ipv4/v6 nexthop
BGP_EXT_TRA_RATE = 0x8006  # traffic-rate 2-byte as#, 4-byte float
BGP_EXT_TRA_ACTION = 0x8007  # traffic-action bitmask
BGP_EXT_REDIRECT_VRF = 0x8008  # redirect 6-byte Route Target
BGP_EXT_TRA_MARK = 0x8009  # traffic-marking DSCP value

# Transitive Opaque
BGP_EXT_COM_OSPF_ROUTE_TYPE = 0x0306  # OSPF Route Type
BGP_EXT_COM_COLOR = 0x030b  # Color
BGP_EXT_COM_COLOR_00 = 0x030b0000  # Color-00
BGP_EXT_COM_COLOR_01 = 0x030b4000  # Color-01
BGP_EXT_COM_COLOR_10 = 0x030b8000  # Color-10
BGP_EXT_COM_COLOR_11 = 0x030bc000  # Color-11
BGP_EXT_COM_ENCAP = 0x030c  # BGP_EXT_COM_ENCAP = 0x030c
BGP_EXT_COM_DEFAULT_GATEWAY = 0x030d  # Default Gateway

# BGP EVPN
BGP_EXT_COM_EVPN_MAC_MOBIL = 0x0600  # Mac Mobility
BGP_EXT_COM_EVPN_ESI_MPLS_LABEL = 0x0601  # ESI MPLS Label
BGP_EXT_COM_EVPN_ES_IMPORT = 0x0602  # ES Import
BGP_EXT_COM_EVPN_ROUTE_MAC = 0x0603  # EVPN Router MAC Extended Community

# BGP cost cummunity
BGP_EXT_COM_COST = 0x4301

# BGP link bandwith
BGP_EXT_COM_LINK_BW = 0x4004

# Unkonw
BGP_EXT_COM_UNKNOW = 0x0000

BGP_EXT_COM_DICT = {
    'redirect-vrf': 32776,  # redirect 6-byte Route Target
    'traffic-marking-dscp': 32777,  # traffic-marking DSCP value
    'traffic-rate': 32774,  # traffic-rate 2-byte as#, 4-byte float
    'color': 779,  # Color
    # Color, leftmost 2 bits of reserved field = 00, CO bits = 00,
    # srpolicy -> IGP
    'color-00': 51052544,
    # Color, leftmost 2 bits of reserved field = 01, CO bits = 01,
    # srpolicy -> same afi null endpoint -> any null endpoint -> IGP
    'color-01': 51068928,
    # Color, leftmost 2 bits of reserved field = 10, CO bits = 10,
    # srpolicy -> same afi null endpoint -> any null endpoint -> same afi endpoint -> any endpoint -> IGP
    'color-10': 51085312,
    # Color, leftmost 2 bits of reserved field = 11, CO bits = 11,
    # treated like color-00
    'color-11': 51101696,
    'encapsulation': 780,  # BGP_EXT_COM_ENCAP = 0x030c
    'es-import': 1538,  # ES Import
    'router-mac': 1539  # EVPN Router MAC Extended Community
}

BGP_EXT_COM_DICT_1 = {
    'esi-label': 1537,  # ESI MPLS Label
    'mac-mobility': 1536,  # Mac Mobility
}

# route distinguisher type
BGP_ROUTE_DISTINGUISHER_TYPE_0 = 0x0000
BGP_ROUTE_DISTINGUISHER_TYPE_1 = 0x0001
BGP_ROUTE_DISTINGUISHER_TYPE_2 = 0x0002

# PMSI TUNNEL TYPE
PMSI_TUNNEL_TYPE_NO_TUNNEL = 0
PMSI_TUNNEL_TYPE_RSVP_TE_P2MP = 1
PMSI_TUNNEL_TYPE_MLDP_P2MP = 2
PMSI_TUNNEL_TYPE_PIM_SSM_TREE = 3
PMSI_TUNNEL_TYPE_PIM_SM_TREE = 4
PMSI_TUNNEL_TYPE_BIDIR_PIM_TREE = 5
PMSI_TUNNEL_TYPE_INGRESS_REPL = 6
PMSI_TUNNEL_TYPE_MLDP_MP2MP = 7

# NLRI type as define in BGP flow spec RFC
BGPNLRI_FSPEC_DST_PFIX = 1  # RFC 5575
BGPNLRI_FSPEC_SRC_PFIX = 2  # RFC 5575
BGPNLRI_FSPEC_IP_PROTO = 3  # RFC 5575
BGPNLRI_FSPEC_PORT = 4  # RFC 5575
BGPNLRI_FSPEC_DST_PORT = 5  # RFC 5575
BGPNLRI_FSPEC_SRC_PORT = 6  # RFC 5575
BGPNLRI_FSPEC_ICMP_TP = 7  # RFC 5575
BGPNLRI_FSPEC_ICMP_CD = 8  # RFC 5575
BGPNLRI_FSPEC_TCP_FLAGS = 9  # RFC 5575
BGPNLRI_FSPEC_PCK_LEN = 10  # RFC 5575
BGPNLRI_FSPEC_DSCP = 11  # RFC 5575
BGPNLRI_FSPEC_FRAGMENT = 12  # RFC 5575

# Sub-TLVs as defined in SR TE Policy draft
BGP_BSID_PREFERENCE_OLD_OR_NEW = 0
BGPSUB_TLV_PREFERENCE = 6
BGPSUB_TLV_BINDGINGSID = 7
BGPSUB_TLV_PREFERENCE_NEW = 12
BGPSUB_TLV_BINDGINGSID_NEW = 13
BGPSUB_TLV_SIDLIST = 128

# Sub-TLVs as defined in SR TE Policy draft and used in BGPSUB_TLV_SIDLIST
BGPSUB_TLV_WEIGHTED = 9
BGPSUB_TLV_SID = 1

# NLRI type as define in BGP EVPN
BGPNLRI_EVPN_ETHERNET_AUTO_DISCOVERY = 1
BGPNLRI_EVPN_MAC_IP_ADVERTISEMENT = 2
BGPNLRI_EVPN_INCLUSIVE_MULTICAST_ETHERNET_TAG = 3
BGPNLRI_EVPN_ETHERNET_SEGMENT = 4
BGPNLRI_EVPN_IP_ROUTE_PREFIX = 5

# BGP message Constants
VERSION = 4
PORT = 179
HDR_LEN = 19
MAX_LEN = 4096

# BGP messages type
MSG_BGP_CLOSED = 0
MSG_OPEN = 1
MSG_UPDATE = 2
MSG_NOTIFICATION = 3
MSG_KEEPALIVE = 4
MSG_ROUTEREFRESH = 5
MSG_CISCOROUTEREFRESH = 128

# BGP Capabilities Support

SUPPORT_4AS = False
CISCO_ROUTE_REFRESH = False
NEW_ROUTE_REFRESH = False
GRACEFUL_RESTART = False

# AFI_SAFI mapping

AFI_SAFI_DICT = {
    (1, 1): 'ipv4',
    (2, 1): 'ipv6',
    (1, 4): 'ipv4_lu',
    (2, 4): 'ipv6_lu',
    (1, 133): 'flowspec',
    (1, 128): 'vpnv4',
    (2, 128): 'vpnv6',
    (25, 70): 'evpn',
    (16388, 71): 'bgpls',
    (1, 73): 'ipv4_srte'
}
AFI_SAFI_STR_DICT = {
    'ipv6': (2, 1),
    'ipv4': (1, 1),
    'ipv4_lu': (1, 4),
    'ipv6_lu': (2, 4),
    'flowspec': (1, 133),
    'vpnv4': (1, 128),
    'vpnv6': (2, 128),
    'evpn': (25, 70),
    'bgpls': (16388, 71),
    'ipv4_srte': (1, 73)
}

ADD_PATH_ACT_DICT = {
    1: 'receive',
    2: 'send',
    3: 'both'
}
# BGP FSM State
ST_IDLE = 1
ST_CONNECT = 2
ST_ACTIVE = 3
ST_OPENSENT = 4
ST_OPENCONFIRM = 5
ST_ESTABLISHED = 6

# BGP Timer (seconds)
DELAY_OPEN_TIME = 10
ROUTE_REFRESH_TIME = 10
LARGER_HOLD_TIME = 4 * 60
CONNECT_RETRY_TIME = 30
IDLEHOLD_TIME = 30
HOLD_TIME = 120

stateDescr = {
    ST_IDLE: "IDLE",
    ST_CONNECT: "CONNECT",
    ST_ACTIVE: "ACTIVE",
    ST_OPENSENT: "OPENSENT",
    ST_OPENCONFIRM: "OPENCONFIRM",
    ST_ESTABLISHED: "ESTABLISHED"
}

# Notification error codes
ERR_MSG_HDR = 1
ERR_MSG_OPEN = 2
ERR_MSG_UPDATE = 3
ERR_HOLD_TIMER_EXPIRED = 4
ERR_FSM = 5
ERR_CEASE = 6
ERR_CAP = 7

# Notification error codes dict
NOTIFICATION_ERROR_CODES_DICT = {
    ERR_MSG_HDR: "Message Header Error",
    ERR_MSG_OPEN: "OPEN Message Error",
    ERR_MSG_UPDATE: "UPDATE Message Error",
    ERR_HOLD_TIMER_EXPIRED: "Hold Timer Expired",
    ERR_FSM: "Finite State Machine Error",
    ERR_CEASE: "Cease",
    ERR_CAP: "CAPABILITY Message Error"
}

# Notification suberror codes for ERR_MSG_HDR
ERR_MSG_HDR_CONN_NOT_SYNC = 1
ERR_MSG_HDR_BAD_MSG_LEN = 2
ERR_MSG_HDR_BAD_MSG_TYPE = 3

# Notification suberror codes for ERR_MSG_OPEN
ERR_MSG_OPEN_UNSUP_VERSION = 1
ERR_MSG_OPEN_BAD_PEER_AS = 2
ERR_MSG_OPEN_BAD_BGP_ID = 3
ERR_MSG_OPEN_UNSUP_OPT_PARAM = 4
ERR_MSG_OPEN_UNACCPT_HOLD_TIME = 6
ERR_MSG_OPEN_UNSUP_CAPA = 7  # RFC 5492
ERR_MSG_OPEN_UNKNO = 8

# Notification suberror codes for ERR_MSG_UPDATE
ERR_MSG_UPDATE_MALFORMED_ATTR_LIST = 1
ERR_MSG_UPDATE_UNRECOGNIZED_WELLKNOWN_ATTR = 2
ERR_MSG_UPDATE_MISSING_WELLKNOWN_ATTR = 3
ERR_MSG_UPDATE_ATTR_FLAGS = 4
ERR_MSG_UPDATE_ATTR_LEN = 5
ERR_MSG_UPDATE_INVALID_ORIGIN = 6
ERR_MSG_UPDATE_INVALID_NEXTHOP = 8
ERR_MSG_UPDATE_OPTIONAL_ATTR = 9
ERR_MSG_UPDATE_INVALID_NETWORK_FIELD = 10
ERR_MSG_UPDATE_MALFORMED_ASPATH = 11
ERR_MSG_UPDATE_UNKOWN_ATTR = 12

# Notification suberror codes for ERR_HOLD_TIMER_EXPIRED
ERR_SUB_HOLD_TIMER_EXPIRED = 1

# Notification suberror codes for ERR_FSM
ERR_SUB_FSM_ERROR = 1

# Notification suberror codes for ERR_CEASE
ERR_MAXIMUM_NUMBER_OF_PREFIXES_REACHED = 1
ERR_ADMINISTRATIVE_SHUTDOWN = 2
ERR_PEER_DECONFIGURED = 3
ERR_ADMINISTRATIVE_RESET = 4
ERR_CONNECTION_RESET = 5
ERR_OTHER_CONFIGURATION_CHANGE = 6
ERR_CONNECTION_COLLISION_RESOLUTION = 7
ERR_OUT_OF_RESOURCES = 8

NOTIFICATION_SUB_ERROR_CODES_DICT = {
    ERR_MSG_HDR: {
        ERR_MSG_HDR_CONN_NOT_SYNC: 'Connection Not Synchronized',  # 1
        ERR_MSG_HDR_BAD_MSG_LEN: 'Bad Message Length',             # 2
        ERR_MSG_HDR_BAD_MSG_TYPE: 'Bad Message Type'               # 3
    },
    ERR_MSG_OPEN: {
        ERR_MSG_OPEN_UNSUP_VERSION: 'Unsupported Version Number',
        ERR_MSG_OPEN_BAD_PEER_AS: 'Bad Peer AS',
        ERR_MSG_OPEN_BAD_BGP_ID: 'Bad BGP Identifier',
        ERR_MSG_OPEN_UNSUP_OPT_PARAM: 'Unsupported Optional Parameter',
        ERR_MSG_OPEN_UNACCPT_HOLD_TIME: 'Unacceptable Hold Time',
        ERR_MSG_OPEN_UNSUP_CAPA: 'Unsupported Capability',
        ERR_MSG_OPEN_UNKNO: 'NULL',
    },
    ERR_MSG_UPDATE: {
        ERR_MSG_UPDATE_MALFORMED_ATTR_LIST: 'Malformed Attribute List',
        ERR_MSG_UPDATE_UNRECOGNIZED_WELLKNOWN_ATTR: 'Unrecognized Well-known Attribute',
        ERR_MSG_UPDATE_MISSING_WELLKNOWN_ATTR: 'Missing Well-known Attribute',
        ERR_MSG_UPDATE_ATTR_FLAGS: 'Attribute Flags Error',
        ERR_MSG_UPDATE_ATTR_LEN: 'Attribute Length Error',
        ERR_MSG_UPDATE_INVALID_ORIGIN: 'Invalid ORIGIN Attribute',
        ERR_MSG_UPDATE_INVALID_NEXTHOP: 'Invalid NEXT_HOP Attribute',
        ERR_MSG_UPDATE_OPTIONAL_ATTR: 'Optional Attribute Error',
        ERR_MSG_UPDATE_INVALID_NETWORK_FIELD: 'Invalid Network Field',
        ERR_MSG_UPDATE_MALFORMED_ASPATH: 'Malformed AS_PATH',
        ERR_MSG_UPDATE_UNKOWN_ATTR: 'NULL'
    },
    ERR_HOLD_TIMER_EXPIRED: {
        ERR_SUB_HOLD_TIMER_EXPIRED: 'Hold timer expired'
    },
    ERR_FSM: {
        ERR_SUB_FSM_ERROR: 'FSM error'
    },
    ERR_CEASE: {
        ERR_MAXIMUM_NUMBER_OF_PREFIXES_REACHED: 'Maximum number of prefixes reached',
        ERR_ADMINISTRATIVE_SHUTDOWN: 'Administrative shutdown',
        ERR_PEER_DECONFIGURED: 'Peer reconfigured',
        ERR_ADMINISTRATIVE_RESET: 'Administrative reset',
        ERR_CONNECTION_RESET: 'Connection reset',
        ERR_OTHER_CONFIGURATION_CHANGE: 'Other configuration change',
        ERR_CONNECTION_COLLISION_RESOLUTION: 'Connection collision resolution',
        ERR_OUT_OF_RESOURCES: 'Out of resources'
    }
}

ATTRIBUTE_ID_2_STR = {
    1: 'ORIGIN',
    2: 'AS_PATH',
    3: 'NEXT_HOP',
    4: 'MULTI_EXIT_DISC',
    5: 'LOCAL_PREF',
    6: 'ATOMIC_AGGREGATE',
    7: 'AGGREGATOR',
    8: 'COMMUNITY',
    9: 'ORIGINATOR_ID',
    10: 'CLUSTER_LIST',
    14: 'MP_REACH_NLRI',
    15: 'MP_UNREACH_NLRI',
    16: 'EXTENDED_COMMUNITY',
    17: 'AS4_PATH',
    18: 'AS4_AGGREGATOR',
    22: 'PMSI_TUNNEL'
}

ATTRIBUTE_STR_2_ID = dict([(v, k) for (k, v) in ATTRIBUTE_ID_2_STR.items()])


WELL_KNOW_COMMUNITY_INT_2_STR = {
    0xFFFF0000: 'PLANNED_SHUT',
    0xFFFF0001: 'ACCEPT_OWN',
    0xFFFF0002: 'ROUTE_FILTER_TRANSLATED_v4',
    0xFFFF0003: 'ROUTE_FILTER_v4',
    0xFFFF0004: 'ROUTE_FILTER_TRANSLATED_v6',
    0xFFFF0005: 'ROUTE_FILTER_v6',
    0xFFFF029A: 'BLACKHOLE',
    0xFFFFFF01: 'NO_EXPORT',
    0xFFFFFF02: 'NO_ADVERTISE',
    0xFFFFFF03: 'NO_EXPORT_SUBCONFED',
    0xFFFFFF04: 'NOPEER'
}

WELL_KNOW_COMMUNITY_STR_2_INT = dict(
    [(r, l) for (l, r) in WELL_KNOW_COMMUNITY_INT_2_STR.items()])

TCP_MD5SIG_MAXKEYLEN = 80
SS_PADSIZE_IPV4 = 120
TCP_MD5SIG = 14
SS_PADSIZE_IPV6 = 100
SIN6_FLOWINFO = 0
SIN6_SCOPE_ID = 0



#!/usr/bin/python
# -*- coding: utf-8 -*-
import abc
import logging
from queue import Queue


LOG = logging.getLogger(__name__)


class BaseHandler(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """
        internal message queue:
        push message in custom handler
        pop message when keep alive received
        """
        self.inter_mq = Queue()
        pass

    @abc.abstractmethod
    def init(self):
        raise NotImplemented

    @abc.abstractmethod
    def on_update_error(self, peer, timestamp, msg):
        raise NotImplemented

    @abc.abstractmethod
    def update_received(self, peer, timestamp, msg):
        raise NotImplemented

    @abc.abstractmethod
    def keepalive_received(self, peer, timestamp):
        raise NotImplemented

    @abc.abstractmethod
    def open_received(self, peer, timestamp, result):
        raise NotImplemented

    @abc.abstractmethod
    def route_refresh_received(self, peer, msg, msg_type):
        raise NotImplemented

    @abc.abstractmethod
    def notification_received(self, peer, msg):
        raise NotImplemented

    @abc.abstractmethod
    def on_connection_lost(self, peer):
        raise NotImplemented

    @abc.abstractmethod
    def on_connection_failed(self, peer, msg):
        raise NotImplemented

    @abc.abstractmethod
    def on_established(self, peer, msg):
        raise NotImplemented



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii

from yabgp.tlv import TLV
from ..linkstate import LinkState


@LinkState.register(_type=1115)
class MinMaxUnidirectLinkDelay(TLV):
    """
     Min/Max Unidirectional Link Delay
    """
    # TYPE = 1115  # https://tools.ietf.org/html/draft-ietf-idr-te-pm-bgp-10#section-3.2
    TYPE_STR = 'min_max_unidirect_link_delay'

    @classmethod
    def unpack(cls, data):
        min_value = int(binascii.b2a_hex(data[:4]), 16)
        max_value = int(binascii.b2a_hex(data[4:]), 16)
        return cls(value={
            "min_link_delay": min_value,
            "max_link_delay": max_value,
        })



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/rfc5307#section-1.2
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |Protection Cap |    Reserved   |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# 0x01  Extra Traffic
# 0x02  Unprotected
# 0x04  Shared
# 0x08  Dedicated 1:1
# 0x10  Dedicated 1+1
# 0x20  Enhanced
# 0x40  Reserved
# 0x80  Reserved


@LinkState.register()
class ProtectionType(TLV):

    TYPE = 1093  # https://tools.ietf.org/html/rfc5307#section-1.2
    TYPE_STR = "protection_type"

    @classmethod
    def unpack(cls, value):
        """
        """
        value = ord(value[0])
        if value == 0x01:
            return cls(value='Extra Traffic')
        if value == 0x02:
            return cls(value='Unprotected')
        if value == 0x04:
            return cls(value='Shared')
        if value == 0x08:
            return cls(value='Dedicated 1:1')
        if value == 0x10:
            return cls(value='Dedicated 1+1')
        if value == 0x20:
            return cls(value='Enhanced')



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""linkstate
"""

import struct
import binascii

import netaddr

from yabgp.message.attribute.nlri import NLRI


class BGPLS(NLRI):
    """BGPLS
    """
    NODE_NLRI = 1
    LINK_NLRI = 2
    IPv4_TOPO_PREFIX_NLRI = 3
    IPv6_TOPO_PREFIX_NLRI = 4

    @classmethod
    def parse(cls, nlri_data):
        nlri_list = []
        while nlri_data:
            _type, length = struct.unpack('!HH', nlri_data[0:4])
            value = nlri_data[4: 4 + length]
            nlri_data = nlri_data[4+length:]
            nlri = dict()
            if _type == cls.LINK_NLRI:
                nlri['type'] = 'link'
            elif _type == cls.NODE_NLRI:
                nlri['type'] = 'node'
            elif _type == cls.IPv4_TOPO_PREFIX_NLRI:
                nlri['type'] = 'ipv4_topo_prefix'
            elif _type == cls.IPv6_TOPO_PREFIX_NLRI:
                nlri['type'] = 'ipv6_topo_prefix'
            else:
                nlri['type'] = 'unknown'
                continue
            protocol_id, identifier, descriptors = cls.parse_nlri(value, _type)
            nlri['protocol_id'] = protocol_id
            nlri['instances_id'] = identifier
            nlri['descriptors'] = descriptors
            nlri_list.append(nlri)
        return nlri_list

    @classmethod
    def parse_nlri(cls, data, nlri_type):
        """parse nlri: node, link, prefix
        """
        #      0                   1                   2                   3
        #      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #     +-+-+-+-+-+-+-+-+
        #     |  Protocol-ID  |
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     |                           Identifier                          |
        #     |                            (64 bits)                          |
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     //               Local Node Descriptors (variable)             //
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     //               Remote Node Descriptors (variable)            //
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     //                  Link Descriptors (variable)                //
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #                     Figure 8: The Link NLRI format

        # The Protocol-ID field can contain one of the following values:
        #     +-------------+----------------------------------+
        #     | Protocol-ID | NLRI information source protocol |
        #     +-------------+----------------------------------+
        #     |      1      | IS-IS Level 1                    |
        #     |      2      | IS-IS Level 2                    |
        #     |      3      | OSPFv2                           |
        #     |      4      | Direct                           |
        #     |      5      | Static configuration             |
        #     |      6      | OSPFv3                           |
        #     |      7      | BGP                              |
        #     +-------------+----------------------------------+
        #               Table 2: Protocol Identifiers

        #     +------------+----------------------------------+
        #     | Identifier | Routing Universe                 |
        #     +------------+----------------------------------+
        #     |     0      | Default Layer 3 Routing topology |
        #     +------------+----------------------------------+
        #         Table 3: Well-Known Instance Identifiers

        proto_id = ord(data[0:1])

        identifier = int(binascii.b2a_hex(data[1:9]), 16)

        descriptor_list = []
        descriptors = data[9:]
        while descriptors:
            _type, length = struct.unpack('!HH', descriptors[0:4])
            value = descriptors[4: 4+length]
            descriptors = descriptors[4+length:]
            descriptor = dict()
            if _type == 256:  # local node
                descriptor['type'] = 'local_node'
                descriptor['value'] = cls.parse_node_descriptor(value, proto_id)

            elif _type == 257:  # remote node
                descriptor['type'] = 'remote_node'
                descriptor['value'] = cls.parse_node_descriptor(value, proto_id)
            # elif _type == 258:  # link local/remote identifier
            #     pass
            elif _type == 259:  # ipv4 interface address
                ipv4_addr = str(netaddr.IPAddress(int(binascii.b2a_hex(value), 16)))
                descriptor['type'] = 'link_local_ipv4'
                descriptor['value'] = ipv4_addr
            elif _type == 260:  # ipv4 neighbor address
                ipv4_neighbor_addr = str(netaddr.IPAddress(int(binascii.b2a_hex(value), 16)))
                descriptor['type'] = 'link_remote_ipv4'
                descriptor['value'] = ipv4_neighbor_addr
            # elif _type == 263: # Multi-Topology Identifier
            #     pass
            elif _type == 263:  # Multi-Topology Identifier
                descriptor['type'] = 'mt_id'
                descriptor['value'] = []
                while value:
                    descriptor['value'].append(struct.unpack('!H', value[:2])[0])
                    value = value[2:]
            elif _type == 264:  # OSPF Route Type
                descriptor['type'] = 'prefix_ospf_route_type'
                descriptor['value'] = ord(value[0:1])
            elif _type == 265:   # IP Reachability Information
                descriptor['type'] = 'prefix'
                mask = ord(value[0:1])
                if nlri_type == cls.IPv4_TOPO_PREFIX_NLRI:
                    prefix_value = value[1:] + b'\x00' * (4 - len(value[1:]))
                    ip_str = str(netaddr.IPAddress(int(binascii.b2a_hex(prefix_value), 16)))
                else:
                    ip_str = str(netaddr.IPAddress(int(binascii.b2a_hex(value[1:]), 16)))
                descriptor['value'] = "%s/%s" % (ip_str, mask)
            else:
                descriptor['type'] = _type
                descriptor['value'] = binascii.b2a_hex(value)
            descriptor_list.append(descriptor)
        return proto_id, identifier, descriptor_list

    @classmethod
    def parse_node_descriptor(cls, data, proto):

        # +--------------------+-------------------+----------+
        # | Sub-TLV Code Point | Description       |   Length |
        # +--------------------+-------------------+----------+
        # |        512         | Autonomous System |        4 |
        # |        513         | BGP-LS Identifier |        4 |
        # |        514         | OSPF Area-ID      |        4 |
        # |        515         | IGP Router-ID     | Variable |
        # |        516         | BGP Router-ID     |        4 |
        # |        517         | Member-ASN        |        4 |
        # +--------------------+-------------------+----------+
        return_data = dict()
        while data:
            _type, length = struct.unpack('!HH', data[0: 4])
            value = data[4: 4+length]
            data = data[4+length:]
            if _type == 512:
                return_data['as_num'] = int(binascii.b2a_hex(value), 16)
            elif _type == 513:
                return_data['bgpls_id'] = str(netaddr.IPAddress(int(binascii.b2a_hex(value), 16)))
            elif _type == 514:
                return_data['ospf_area_id'] = str(netaddr.IPAddress(int(binascii.b2a_hex(value), 16)))
            elif _type == 515:
                # OSPFv2, OSPFv3 non-pseudonode
                if (proto == 3 or proto == 6) and length == 4:
                    return_data['igp_router_id'] = {
                        "pseudonode": False,
                        "router_id": str(netaddr.IPAddress(int(binascii.b2a_hex(value), 16)))
                    }
                # OSPFv2, OSPFv3, LAN pseudonode
                if (proto == 3 or proto == 6) and length == 8:
                    return_data['igp_router_id'] = {
                        "pseudonode": True,
                        "router_id": str(netaddr.IPAddress(int(binascii.b2a_hex(value[:4]), 16))),
                        "designated_router_addr": str(netaddr.IPAddress(int(binascii.b2a_hex(value[4:]), 16)))
                    }
                # IS-IS non-pseudonode
                if (proto == 1 or proto == 2) and length == 6:
                    return_data['igp_router_id'] = {
                        "pseudonode": False,
                        "iso_node_id": cls.parse_iso_node_id(value)
                    }
                # IS-IS LAN pseudonode = ISO Node-ID + PSN
                # Unpack ISO address
                if (proto == 1 or proto == 2) and length == 7:
                    return_data['igp_router_id'] = {
                        "pseudonode": True,
                        "psn": ord(value[6: 7]),
                        "iso_node_id": cls.parse_iso_node_id(value[:6])
                    }
            elif _type == 516:
                return_data['bgp_router_id'] = str(netaddr.IPAddress(int(binascii.b2a_hex(value[:4]), 16)))
            elif _type == 517:
                return_data['member_as_num'] = int(binascii.b2a_hex(value), 16)
        return return_data

    @classmethod
    def parse_iso_node_id(cls, data):
        tmp = binascii.b2a_hex(data)
        chunks, chunk_size = len(tmp), len(tmp)/3
        return '.'.join([tmp[i:i+chunk_size] for i in range(0, chunks, chunk_size)])



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class ByteFlag(int):

    FLAGS = ['']

    def dict(self):
        bit_list = []
        for i in range(8):
            bit_list.append((self >> i) & 1)
        bit_list.reverse()
        return dict(zip(self.FLAGS, bit_list[:len(self.FLAGS)]))



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii


class TLV(object):
    """TLV basic class
    """
    TYPE = -1
    TYPE_STR = "UNKNOWN"

    __slots__ = ['value']

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return '%s: %s' % (self.TYPE_STR, self.value)

    @classmethod
    def unpack(cls, value):
        """unpack binary data
        """
        return cls(value=binascii.b2a_hex(value))

    def dict(self, str_key=True):
        """return dict format
        """
        if str_key:
            return {
                'type': self.TYPE_STR,
                'value': self.value}
        return {
            'type': self.TYPE,
            'value': self.value
        }



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test IPv6 Unicast """

import unittest

from yabgp.message.attribute.nlri.ipv6_unicast import IPv6Unicast


class TestIPv6Unicast(unittest.TestCase):
    def test_parse(self):
        nlri_data = b'\x80\x20\x01\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x40\x20\x01\x32' \
                    b'\x32\x00\x01\x00\x00\x7f\x20\x01\x48\x37\x16\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'
        value_parsed = IPv6Unicast().parse(nlri_data)
        value_hoped = ['2001:3232::1/128', '::2001:3232:1:0/64', '2001:4837:1632::2/127']
        self.assertEqual(value_hoped, value_parsed)

    def test_parse_withdraw(self):
        nlri_data = b'\x40\x20\x01\x0d\xb8\x00\x01\x00\x02\x40\x20\x01\x0d\xb8\x00\x01\x00\x01\x40\x20\x01\x0d' \
                    b'\xb8\x00\x01\x00\x00'
        value_parsed = IPv6Unicast.parse(nlri_data)
        value_hoped = ['::2001:db8:1:2/64', '::2001:db8:1:1/64', '::2001:db8:1:0/64']
        self.assertEqual(value_hoped, value_parsed)

    def test_construct_nlri(self):
        nlri_list = ['::2001:db8:1:2/64', '::2001:db8:1:1/64', '::2001:db8:1:0/64']
        value_hex = IPv6Unicast.construct(nlri_list)
        value_hoped = b'\x40\x20\x01\x0d\xb8\x00\x01\x00\x02\x40\x20\x01\x0d\xb8\x00\x01\x00\x01\x40\x20\x01\x0d' \
                      b'\xb8\x00\x01\x00\x00'
        self.assertEqual(value_hoped, value_hex)

    def test_construct_nlri_2(self):
        nlri_list = ['2001:3232::1/128', '::2001:3232:1:0/64', '2001:4837:1632::2/127']
        value_hex = IPv6Unicast.construct(nlri_list)
        value_hoped = b'\x80\x20\x01\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x40\x20\x01\x32' \
                      b'\x32\x00\x01\x00\x00\x7f\x20\x01\x48\x37\x16\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'
        self.assertEqual(value_hoped, value_hex)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
import json
import os
import time
import logging
import traceback
import sys

from oslo_config import cfg

from yabgp.common import constants as bgp_cons
from yabgp.handler import BaseHandler

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


MSG_PROCESS_OPTS = [
    cfg.BoolOpt('write_disk',
                default=True,
                help='Whether the BGP message is written to disk'),
    cfg.StrOpt('write_dir',
               default=os.path.join(os.environ['HOME'], 'data/bgp/'),
               help='The BGP messages storage path'),
    cfg.IntOpt('write_msg_max_size',
               default=500,
               help='The Max size of one BGP message file, the unit is MB'),
    cfg.BoolOpt('write_keepalive',
                default=False,
                help='Whether write keepalive message to disk')
]

CONF.register_opts(MSG_PROCESS_OPTS, group='message')


class DefaultHandler(BaseHandler):
    def __init__(self):
        super(DefaultHandler, self).__init__()
        '''
            {<peer>: (<path>, <current file>)}
        '''
        self.peer_files = {}
        '''
            {<peer>: <seq number>}
        '''
        self.msg_sequence = {}

    def init(self):
        if CONF.message.write_disk:
            self.init_msg_file(CONF.bgp.running_config['remote_addr'].lower())

    def init_msg_file(self, peer_addr):
        msg_file_path_for_peer = os.path.join(
            CONF.message.write_dir,
            peer_addr
        )
        if not os.path.exists(msg_file_path_for_peer):
            os.makedirs(msg_file_path_for_peer)
            LOG.info('Create dir %s for peer %s', msg_file_path_for_peer, peer_addr)
        LOG.info('BGP message file path is %s', msg_file_path_for_peer)

        if msg_file_path_for_peer and peer_addr not in self.peer_files:
            msg_path = msg_file_path_for_peer + '/msg/'
            if not os.path.exists(msg_path):
                os.makedirs(msg_path)

            # try get latest file and msg sequence if any
            last_msg_seq, msg_file_name = DefaultHandler.get_last_seq_and_file(msg_path)

            if not msg_file_name:
                msg_file_name = "%s.msg" % time.time()
            # store the message sequence
            self.msg_sequence[peer_addr] = last_msg_seq + 1
            msg_file = open(os.path.join(msg_path, msg_file_name), 'a')
            msg_file.flush()
            self.peer_files[peer_addr] = (msg_path, msg_file)
            LOG.info('BGP message file %s', msg_file_name)
            LOG.info('The last bgp message seq number is %s', last_msg_seq)

    @staticmethod
    def get_last_seq_and_file(msg_path):
        """
        Get the last sequence number in the latest log file.
        """
        LOG.info('get the last bgp message seq for this peer')
        last_seq = 0
        # first get the last file
        file_list = os.listdir(msg_path)
        if not file_list:
            return last_seq, None
        file_list.sort()
        msg_file_name = file_list[-1]
        try:
            with open(msg_path + msg_file_name, 'r') as fh:
                line = None
                for line in fh:
                    pass
                last = line
                if line:
                    if last.startswith('['):
                        last_seq = eval(last)[1]
                    elif last.startswith('{'):
                        last_seq = json.loads(last)['seq']
        except OSError:
            LOG.error('Error when reading bgp message files')
        except Exception as e:
            LOG.debug(traceback.format_exc())
            LOG.error(e)
            sys.exit()

        return last_seq, msg_file_name

    def write_msg(self, peer, timestamp, msg_type, msg):
        """
        write bgp message into local disk file
        :param peer: peer address
        :param timestamp: timestamp
        :param msg_type: message type (0,1,2,3,4,5,6)
        :param msg: message dict
        :param msg_path: path to store messages on disk
        :return:
        """
        msg_path, msg_file = self.peer_files.get(peer.lower(), (None, None))
        if msg_path:
            msg_seq = self.msg_sequence[peer.lower()]

            msg_record = {
                't': timestamp,
                'seq': msg_seq,
                'type': msg_type
            }
            msg_record.update(msg)
            try:
                json.dump(msg_record, msg_file)
            except Exception as e:
                LOG.error(e)
                LOG.info('raw message %s', msg)
            msg_file.write('\n')
            self.msg_sequence[peer.lower()] += 1
            msg_file.flush()
            os.fsync(msg_file.fileno())

    def check_file_size(self, peer):
        """if the size of the msg file is bigger than 'max_msg_file_size',
        then save as and re-open a new file.
        """
        msg_path, cur_file = self.peer_files.get(peer.lower(), (None, None))
        if msg_path:
            if os.path.getsize(cur_file.name) >= CONF.message.write_msg_max_size:
                cur_file.close()
                msg_file_name = "%s.msg" % time.time()
                LOG.info('Open a new message file %s', msg_file_name)
                msg_file = open(os.path.join(msg_path + msg_file_name), 'a')
                self.peer_files[peer.lower()] = (msg_path, msg_file)
                return True
        return False

    def on_update_error(self, peer, timestamp, msg):
        self.write_msg(
            peer=peer.factory.peer_addr,
            timestamp=timestamp,
            msg_type=6,
            msg={'msg': msg}
        )

    def update_received(self, peer, timestamp, msg):
        # write message to disk
        self.write_msg(
            peer=peer.factory.peer_addr,
            timestamp=timestamp,
            msg_type=bgp_cons.MSG_UPDATE,
            msg={"msg": msg}
        )
        self.check_file_size(peer.factory.peer_addr)

    def keepalive_received(self, peer, timestamp):
        """
        keepalive message default handler
        :param peer:
        :param timestamp:
        :return:
        """

        if peer.msg_recv_stat['Keepalives'] == 1:
            # do something with the connection establish event
            pass

        if CONF.message.write_keepalive:
            # write bgp message
            self.write_msg(
                peer=peer.factory.peer_addr,
                timestamp=timestamp,
                msg_type=4,
                msg={"msg": None}
            )

    def open_received(self, peer, timestamp, result):
        # write bgp message
        self.write_msg(
            peer=peer.factory.peer_addr,
            timestamp=timestamp,
            msg_type=1,
            msg={"msg": result}
        )

    def route_refresh_received(self, peer, msg, msg_type):
        self.write_msg(
            peer=peer.factory.peer_addr,
            timestamp=time.time(),
            msg_type=msg_type,
            msg={"msg": msg}
        )

    def notification_received(self, peer, msg):
        self.write_msg(
            peer=peer.factory.peer_addr,
            timestamp=time.time(),
            msg_type=3,
            msg={"msg": msg}
        )

    def on_connection_lost(self, peer):
        self.write_msg(
            peer=peer.factory.peer_addr,
            timestamp=time.time(),
            msg_type=bgp_cons.MSG_BGP_CLOSED,
            msg={"msg": None}
        )

    def on_connection_failed(self, peer, msg):
        self.write_msg(
            peer=peer,
            timestamp=time.time(),
            msg_type=0,
            msg={"msg": msg}
        )

    def on_established(self, peer, msg):
        pass



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import os
import logging
import traceback

from oslo_config import cfg
from twisted.internet import reactor
from twisted.web.server import Site
from twisted.web.wsgi import WSGIResource

from yabgp import version, log
from yabgp.core.factory import BGPPeering
from yabgp.config import get_bgp_config
from yabgp.common import constants as bgp_cons
from yabgp.api.app import app
from yabgp.handler.default_handler import DefaultHandler


log.early_init_log(logging.DEBUG)

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


def check_msg_config():
    """check message configuration
    """
    LOG.info('Check configurations about message process')
    if CONF.message.write_disk:
        if not os.path.exists(CONF.message.write_dir):
            try:
                os.makedirs(CONF.message.write_dir)
                LOG.info('Create dir %s', CONF.message.write_dir)
            except OSError:
                # sencond try
                if not os.path.exists(CONF.message.write_dir):
                    LOG.error(traceback.format_exc())
                    sys.exit()
        CONF.message.write_msg_max_size = CONF.message.write_msg_max_size * 1024 * 1024


def prepare_twisted_service(handler, reactor_thread_size=100):
    """prepare twsited service
    """
    LOG.info('Prepare twisted services')

    LOG.info('Get peer configuration')
    for conf_key in CONF.bgp.running_config:
        LOG.info('---%s = %s', conf_key, CONF.bgp.running_config[conf_key])

    # init handler
    handler.init()

    LOG.info('Create BGPPeering twsited instance')
    afi_safi_list = [bgp_cons.AFI_SAFI_STR_DICT[afi_safi] for afi_safi in CONF.bgp.running_config['afi_safi']]
    CONF.bgp.running_config['afi_safi'] = afi_safi_list
    CONF.bgp.running_config['capability']['local']['afi_safi'] = afi_safi_list
    bgp_peering = BGPPeering(
        myasn=CONF.bgp.running_config['local_as'],
        myaddr=CONF.bgp.running_config['local_addr'],
        peerasn=CONF.bgp.running_config['remote_as'],
        peeraddr=CONF.bgp.running_config['remote_addr'],
        afisafi=CONF.bgp.running_config['afi_safi'],
        md5=CONF.bgp.running_config['md5'],
        handler=handler
    )
    CONF.bgp.running_config['factory'] = bgp_peering

    # Starting api server
    LOG.info("Prepare RESTAPI service")
    LOG.info("reactor_thread_size = %s", reactor_thread_size)
    reactor.suggestThreadPoolSize(reactor_thread_size)
    resource = WSGIResource(reactor, reactor.getThreadPool(), app)
    site = Site(resource)
    try:
        reactor.listenTCP(CONF.rest.bind_port, site, interface=CONF.rest.bind_host)
        LOG.info("serving RESTAPI on http://%s:%s", CONF.rest.bind_host, CONF.rest.bind_port)
    except Exception as e:
        LOG.error(e, exc_info=True)
        sys.exit()

    LOG.info('Starting BGPPeering twsited instance')
    bgp_peering.automatic_start()

    reactor.run()


def register_api_handler(api_handler):
    """register flask blueprint
    """
    app.register_blueprint(api_handler.blueprint, url_prefix=api_handler.url_prefix)


def prepare_service(args=None, handler=None, api_hander=None, reactor_thread_size=100):
    """prepare all services
    """
    try:
        CONF(args=args, project='yabgp', version=version,
             default_config_files=['/etc/yabgp/yabgp.ini'])
    except cfg.ConfigFilesNotFoundError:
        CONF(args=args, project='yabgp', version=version)

    log.init_log()
    LOG.info('Log (Re)opened.')
    LOG.info("Configuration:")
    cfg.CONF.log_opt_values(LOG, logging.INFO)
    try:
        if not handler:
            LOG.info('No handler provided, init default handler')
            handler = DefaultHandler()
        get_bgp_config()
        check_msg_config()
    except Exception as e:
        LOG.error(e)
        LOG.debug(traceback.format_exc())
        sys.exit()
    # prepare api handler
    if api_hander:
        register_api_handler(api_hander)
    LOG.info('Starting server in PID %s', os.getpid())
    prepare_twisted_service(handler, reactor_thread_size)



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.tlv import TLV
from ..linkstate import LinkState

#     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |            Type               |            Length             |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |  Algorithm 1  |  Algorithm... |  Algorithm N |                |
#    +-                                                             -+
#    |                                                               |
#    +                                                               +


@LinkState.register()
class SRAlgorithm(TLV):
    """
    sr algorithm
    """
    TYPE = 1035  # https://tools.ietf.org/html/draft-ietf-idr-bgp-ls-segment-routing-ext-03#section-2.1.3
    TYPE_STR = 'sr_algorithm'

    @classmethod
    def unpack(cls, data):
        if type(data) is str:
            results = [struct.unpack('!B', value)[0] for value in data]
        else:
            results = list(data)
        return cls(value=results)
